/* global describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const {expect} = require('chai')
const Expression = require('./expression')

describe('Expression', function () {
  describe('variables', function () {
    it('should pull out variables', function () {
      _.each({
        'O1&&P1': ['O1', 'P1'],
        '(G1&&G2)||(G3&&G4)': ['G1', 'G2', 'G3', 'G4']
      }, function (expected, expression) {
        var variables = Expression.variables(expression)
        expect(variables).to.eql(expected)
      })
    })
  })

  describe('evaluate', function () {
    it('should not evaluate invalid expressions', function () {
      expect(Expression.evaluate(undefined, ['O1'])).to.be.false
      expect(Expression.evaluate('', ['O1'])).to.be.false
    })

    it('should evaluate valid expressions', function () {
      expect(Expression.evaluate('false', ['O1', 'P1', 'P2'])).to.be.false
      expect(Expression.evaluate('true', ['O1', 'P1', 'P2'])).to.be.true
      expect(Expression.evaluate('O1&&P1', ['O1', 'P1', 'P2'])).to.be.true
      expect(Expression.evaluate('O1&&P1', ['O1'])).to.be.false
      expect(Expression.evaluate('(G1&&G2)||(G3&&G4)', ['G1', 'G2'])).to.be.true
      expect(Expression.evaluate('(G1&&G2)||(G3&&G4)', ['G1', 'G3'])).to.be.false
      expect(Expression.evaluate('O1&&!(P2||P3)', ['O1', 'P1'])).to.be.true
      expect(Expression.evaluate('O1&&!(P2||P3)', ['O1', 'P2'])).to.be.false
    })
  })

  describe('replaceVariable', function () {
    it('should replace variables', function () {
      expect(Expression.replaceVariable('O1||M1', 'M1', '')).to.eql('O1||')
      expect(Expression.replaceVariable('O1||P1', 'P1', 'P2')).to.eql('O1||P2')
      expect(Expression.replaceVariable('O1||P1||O2||P11', 'P1', 'P2')).to.eql('O1||P2||O2||P11')
    })
  })

  describe('removeVariable', function () {
    it('should remove variables', function () {
      expect(Expression.removeVariable('P1', 'P1')).to.be.empty
      expect(Expression.removeVariable('O1||P1', 'O1')).to.eql('P1')
      expect(Expression.removeVariable('O1||P1', 'P1')).to.eql('O1')
      expect(Expression.removeVariable('O1&&!P1', 'O1')).to.eql('!P1')
      expect(Expression.removeVariable('O1&&!P1', 'P1')).to.eql('O1')
      expect(Expression.removeVariable('O1||P1||O2||P11', 'P1')).to.eql('O1||O2||P11')
      expect(Expression.removeVariable('O1||P1||O2||P11', 'P11')).to.eql('O1||P1||O2')
      expect(Expression.removeVariable('(O1||P1)&&!O2', 'P1')).to.eql('O1&&!O2')
      expect(Expression.removeVariable('(O1||P1)&&!O2', 'O2')).to.eql('O1||P1')
      expect(Expression.removeVariable('(O1||P1)&&!O2&&!O3', 'P1')).to.eql('O1&&!O2&&!O3')
      expect(Expression.removeVariable('(O1||P1)&&!O2&&!O3', 'O2')).to.eql('(O1||P1)&&!O3')
      expect(Expression.removeVariable('O1&&!(O2||P1||O3)', 'P1')).to.eql('O1&&!(O2||O3)')
      expect(Expression.removeVariable('O1&&!(O2||P1||O3)', 'O1')).to.eql('!(O2||P1||O3)')
      expect(Expression.removeVariable('(O1&&P1)||(O2&&P2)||(O3&&P3)||O1', 'O1')).to.eql('P1||(O2&&P2)||(O3&&P3)')
      expect(Expression.removeVariable('(O1&&P1)||(O2&&P2)||(O3&&P3)||O1', 'P1')).to.eql('O1||(O2&&P2)||(O3&&P3)||O1')
    })
  })

  describe('Basic', function () {
    it('should decode/encode an expression', function () {
      let expression = Expression.Basic.parse('G2||O2&&!P55')
      let includes = _.invoke(expression.includes, 'toString')
      let excludes = _.invoke(expression.excludes, 'toString')
      expect(includes).to.eql(['G2', 'O2'])
      expect(excludes).to.eql(['P55'])
      expect(expression.toString()).to.eql('(G2||O2)&&!P55')

      expression = Expression.Basic.parse('P10||G3')
      includes = _.invoke(expression.includes, 'toString')
      excludes = _.invoke(expression.excludes, 'toString')
      expect(includes).to.eql(['P10', 'G3'])
      expect(excludes).to.eql([])
      expect(expression.toString()).to.eql('P10||G3')

      expression = Expression.Basic.parse('(P11||P13)&&!(P14||P8)')
      includes = _.invoke(expression.includes, 'toString')
      excludes = _.invoke(expression.excludes, 'toString')
      expect(includes).to.eql(['P11', 'P13'])
      expect(excludes).to.eql(['P14', 'P8'])
      expect(expression.toString()).to.eql('(P11||P13)&&!(P14||P8)')
    })
  })

  describe('Organic', function () {
    describe('parse', function () {
      it('throw on invalid expressions', function () {
        expect(Expression.Organic.parse.bind(Expression.Organic, 'O1||O2')).to.throw('Invalid Expression')
        expect(Expression.Organic.parse.bind(Expression.Organic, 'O1&&O2')).to.throw('Invalid Expression')
        expect(Expression.Organic.parse.bind(Expression.Organic, 'G1&&(G2||G3)')).to.throw('Invalid Expression')
      })

      describe('no organization', function () {
        it('should parse no groups', function () {
          var expression = Expression.Organic.parse(null)
          expect(expression.organizationId).to.be.undefined
          var includes = _.invoke(expression.includes, 'toString')
          expect(includes).to.eql([])
          expect(expression.toString()).to.eql(null)
        })

        it('should parse 1 group', function () {
          var expression = Expression.Organic.parse('G1')
          expect(expression.organizationId).to.be.undefined
          var includes = _.invoke(expression.includes, 'toString')
          expect(includes).to.eql(['G1'])
          expect(expression.toString()).to.eql('G1')
        })

        it('should parse 2 groups', function () {
          var expression = Expression.Organic.parse('G1||G2')
          expect(expression.organizationId).to.be.undefined
          var includes = _.invoke(expression.includes, 'toString')
          expect(includes).to.eql(['G1', 'G2'])
          expect(expression.toString()).to.eql('G1||G2')
        })
      })

      describe('with organization', function () {
        it('should parse no groups', function () {
          var expression = Expression.Organic.parse('O1')
          expect(expression.organizationId).to.eql(1)
          var includes = _.invoke(expression.includes, 'toString')
          expect(includes).to.eql([])
          expect(expression.toString()).to.eql('O1')
        })

        it('should parse 1 group', function () {
          var expression = Expression.Organic.parse('O1&&G1')
          expect(expression.organizationId).to.eql(1)
          var includes = _.invoke(expression.includes, 'toString')
          expect(includes).to.eql(['G1'])
          expect(expression.toString()).to.eql('O1&&G1')
        })

        it('should parse 2 groups', function () {
          var expression = Expression.Organic.parse('O1&&(G1||G2)')
          expect(expression.organizationId).to.eql(1)
          var includes = _.invoke(expression.includes, 'toString')
          expect(includes).to.eql(['G1', 'G2'])
          expect(expression.toString()).to.eql('O1&&(G1||G2)')
        })
      })
    })

    describe('isSubset', function () {
      describe('no organization', function () {
        it('should calculate with no groups', function () {
          expect(Expression.Organic.isSubsetOf(null, null)).to.be.true
          expect(Expression.Organic.isSubsetOf('G1', null)).to.be.true
          expect(Expression.Organic.isSubsetOf('G1||G2', null)).to.be.true
          expect(Expression.Organic.isSubsetOf('O1', null)).to.be.true
          expect(Expression.Organic.isSubsetOf('O1&&G1', null)).to.be.true
          expect(Expression.Organic.isSubsetOf('O1&&(G1||G2)', null)).to.be.true
        })

        it('should calculate with 1 group', function () {
          expect(Expression.Organic.isSubsetOf(null, 'G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1', 'G1')).to.be.true
          expect(Expression.Organic.isSubsetOf('G1||G2', 'G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1', 'G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1&&G1', 'G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1&&(G1||G2)', 'G1')).to.be.false
        })

        it('should calculate with 2 groups', function () {
          expect(Expression.Organic.isSubsetOf(null, 'G1||G2')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1', 'G1||G2')).to.be.true
          expect(Expression.Organic.isSubsetOf('G1||G2', 'G1||G2')).to.be.true
          expect(Expression.Organic.isSubsetOf('O1', 'G1||G2')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1&&G1', 'G1||G2')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1&&(G1||G2)', 'G1||G2')).to.be.false
        })
      })

      describe('with organization', function () {
        it('should calculate with no groups', function () {
          expect(Expression.Organic.isSubsetOf(null, 'O1')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1', 'O1')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1||G2', 'O1')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1', 'O1')).to.be.true
          expect(Expression.Organic.isSubsetOf('O1&&G1', 'O1')).to.be.true
          expect(Expression.Organic.isSubsetOf('O1&&(G1||G2)', 'O1')).to.be.true
        })

        it('should calculate with 1 group', function () {
          expect(Expression.Organic.isSubsetOf(null, 'O1&&G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1', 'O1&&G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1||G2', 'O1&&G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1', 'O1&&G1')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1&&G1', 'O1&&G1')).to.be.true
          expect(Expression.Organic.isSubsetOf('O1&&(G1||G2)', 'O1&&G1')).to.be.false
        })

        it('should calculate with 2 groups', function () {
          expect(Expression.Organic.isSubsetOf(null, 'O1&&(G1||G2)')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1', 'O1&&(G1||G2)')).to.be.false
          expect(Expression.Organic.isSubsetOf('G1||G2', 'O1&&(G1||G2)')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1', 'O1&&(G1||G2)')).to.be.false
          expect(Expression.Organic.isSubsetOf('O1&&G1', 'O1&&(G1||G2)')).to.be.true
          expect(Expression.Organic.isSubsetOf('O1&&(G1||G2)', 'O1&&(G1||G2)')).to.be.true
        })
      })
    })
  })
})
