const _ = require('underscore')

// Map of type to variant parameters for that type
var types = {
  organization: {
    abbr: 'O',
    iconCls: 'fa-building-o',
    name: function (organization) { return organization.name }
  },
  group: {
    abbr: 'G',
    iconCls: 'fa-users',
    name: function (group) { return group.name }
  },
  person: {
    abbr: 'P',
    iconCls: 'fa-user',
    name: function (person) { return person.fullName || person.email }
  }
}

// Map of abbreviations to their type
var abbreviations = _.reduce(types, function (abbrs, params, type) {
  abbrs[params.abbr] = type
  return abbrs
}, {})

// Constructor
var Variant = function (type, id, object) {
  this.type = type
  this.id = id
  this.object = object

  var params = _.omit(types[type], 'name')
  _.extend(this, params)
}

_.extend(Variant, {
  types: types,

  fromString: function (value) {
    if (!value) return
    var matches = value.match(/(\w)(\d+)/)
    if (!matches) return

    var abbr = matches[1]
    var id = parseInt(matches[2], 10)
    var type = abbreviations[abbr]
    if (type) return new Variant(type, id)
  },

  fromOrganization: function (organization) {
    return new Variant('organization', organization.id, organization)
  },

  fromGroup: function (group) {
    return new Variant('group', group.id, group)
  },

  fromPerson: function (person) {
    return new Variant('person', person.id, person)
  },

  clone: function (variant) {
    return new Variant(variant.type, variant.id, variant.object)
  },

  // Determine if these two variants are equal
  isEqual: function (v1, v2) {
    return v1.toString() === v2.toString()
  },

  // Determine if this array of variants contains the variant
  contains: function (variants, variant) {
    return _.any(variants, function (v) {
      return Variant.isEqual(v, variant)
    })
  }
})

_.extend(Variant.prototype, {
  toString: function () {
    return this.uid
  }
})

Object.defineProperty(Variant.prototype, 'uid', {
  enumerable: true,
  get: function () {
    return this.abbr + this.id
  }
})

Object.defineProperty(Variant.prototype, 'name', {
  enumerable: true,
  get: function () {
    if (this.object) return types[this.type].name(this.object)
  }
})

Object.defineProperty(Variant.prototype, 'email', {
  enumerable: true,
  get: function () {
    if (this.object) return this.object.email
  }
})

module.exports = Variant
