const Promise = require('bluebird')
const httpError = require('http-errors')

module.exports = function (Rbac) {
  const {Group, Organization, Person} = Rbac

  function loadById (req, res, next, id) {
    if (!req.currentUser) return next(httpError.Unauthorized())
    Person.findById(id).whereUserCan(req.currentUser).then(function (person) {
      if (!person) return next(httpError.NotFound('Person Not Found'))
      req.person = person
      next()
    }).catch(next)
  }

  return {
    loadById: loadById,

    loadByQuery: function (req, res, next) {
      var personId = parseInt(req.query.person)
      personId ? loadById(req, res, next, personId) : next()
    },

    canIndex: function (req, res, next) {
      if (req.currentUser.admin) return next()

      // Parse query
      var groupId = parseInt(req.query.group, 10)
      var organizationId = parseInt(req.query.organization, 10)
      if (req.organization) organizationId = req.organization.id

      // Try to see if we are a member of the organization/group (those specified)
      Promise.all([
        groupId ? Group.findById(groupId).whereUserCan(req.currentUser) : null,
        organizationId ? Organization.findById(organizationId).whereUserCan(req.currentUser) : null
      ]).spread(function (group, organization) {
        if (group || organization) return next()
        next(httpError.Forbidden('Access Denied'))
      })
    },

    index: function (req, res, next) {
      var qb = Person.query().joinEmail().whereQuery(req.query)
      if (!req.currentUser.admin) qb = qb.whereUserCan(req.currentUser)
      if (req.organization) qb = qb.whereOrganizationId(req.organization.id)
      qb.then(function (people) {
        return Person.loadGroupIds(people)
      }).then(function (people) {
        res.json(people)
      }).catch(next)
    },

    create: function (req, res, next) {
      Person.bulkCreate(req.organization, req.body).then(function (people) {
        return Person.loadGroupIds(people)
      }).then(function (people) {
        res.json(people)
      }).catch(next)
    },

    patch: function (req, res, next) {
      Person.updateBy(req.person, req.body).then(function (person) {
        res.json(person)
      }).catch(next)
    },

    destroy: function (req, res, next) {
      Person.purge(req.person).then(function () {
        res.json(null)
      }).catch(next)
    },

    download: function (req, res, next) {
      // Create stream
      var Readable = require('stream').Readable
      var stream = new Readable()
      Person.bulkDownload(req.organization, stream).then(function () {
        res.setHeader('Content-Disposition', 'attachment; filename=members.csv')
        res.setHeader('Content-Type', 'text/csv')
        stream.pipe(res)
      }).catch(next)
    },

    upload: function (req, res, next) {
      req.busboy.on('file', function (fieldname, file, filename) {
        Person.bulkUpload(req.organization, file, req.currentUser).then(function (people) {
          return Person.loadGroupIds(people)
        }).then(function (people) {
          res.json(people)
        }).catch(next)
      })

      req.pipe(req.busboy)
    }
  }
}
