const _ = require('underscore')
const Promise = require('bluebird')
const moment = require('moment')
const csv = require('fast-csv')
const util = require('util')

// Column schema
// name: Name of the column
// field: key to enter in the row object
// regEx: regular expression to search for in the header
// required: whether or not this column is required
// unique: whether or not this column must be unique across rows
// type: type of column (e.g. array, date)
// exclude: don't allow these values

// Custom CSV Error
function CsvError (message) {
  this.name = 'CsvError'
  this.message = (message || '')
}

util.inherits(CsvError, Error)

// Map headers to columns
function mapHeadersToColumns (headers, schema) {
  var lastColumn
  return _.map(headers, function (header) {
    lastColumn = header && header.length ? _.find(schema, function (entry) {
      return header.match(entry.regEx)
    }) : lastColumn
    return lastColumn
  })
}

// Find the column at this index (Allow past last column, if array)
function findColumnAtIndex (columns, index) {
  if (index < columns.length) return columns[index]
  var lastColumn = _.last(columns)
  if (lastColumn && lastColumn.type === 'array') return lastColumn
}

// Allow
function allowValue (value, column) {
  // Valid?
  var valid = column && value && value.length
  if (!valid) return false

  // Excluded?
  var exclude = _.any(column.exclude, function (exclude) {
    return exclude.toLowerCase() === value.toLowerCase()
  })
  return !exclude
}

// Map a format to date
function mapValueToDate (value, dateFormat) {
  var date = moment(value, dateFormat, true)
  if (date.isValid()) return date.toDate()
}

// Dates
var dateFormats = ['YYYY-MM-DD', 'M/D/YYYY']
function parseDate (value, column) {
  // Find a format that works
  var date
  var index = 0
  while (index < dateFormats.length && !date) {
    date = mapValueToDate(value, dateFormats[index])
    index++
  }

  // Found date? great.
  if (date) return date

  // Throw error, if invalid
  var error = column.name + ' must have valid format (' + dateFormats.join(' or ') + '). Found invalid value: ' + value + '.'
  throw new CsvError(error)
}

// Parse the CSV row
function parseRow (row, columns) {
  // Initialize the result
  var result = _.chain(columns).compact().reduce(function (result, column) {
    if (column.type === 'array') result[column.field] = []
    return result
  }, {}).value()

  // Go through the rows
  return _.reduce(row, function (result, value, index) {
    // Determine if this value is allowed
    var column = findColumnAtIndex(columns, index)
    if (!allowValue(value, column)) return result

    // Check formatting by type
    if (column.type === 'date') {
      value = parseDate(value, column)
    } else if (column.type === 'email') {
      value = value.toLowerCase()
    }

    // Add value
    if (column.type === 'array') {
      result[column.field].push(value)
    } else {
      result[column.field] = value
    }

    return result
  }, result)
}

// Valid the resulting data
function validateColumns (columns, schema) {
  // Check required columns
  var required = _.where(schema, {required: true})
  _.each(required, function (column) {
    var exists = _.contains(columns, column)
    if (!exists) throw new CsvError('Missing required column: ' + column.name)
  })
}

// Valid the resulting data
function validateRows (rows, schema) {
  // Check required columns
  var required = _.where(schema, {required: true})
  _.each(required, function (column) {
    _.each(rows, function (row, index) {
      var value = row[column.field]
      if (!value) throw new CsvError('Missing ' + column.name + ' on line ' + (index + 2))
    })
  })

  // Check unique columns
  var unique = _.where(schema, {unique: true})
  _.each(unique, function (column) {
    // Search for a duplicate
    var duplicate = _.chain(rows).pluck(column.field).compact().countBy().reduce(function (result, count, value) {
      return result || (count > 1 ? value : null)
    }, null).value()

    // If duplicate, throw error
    if (duplicate) throw new CsvError('Duplicate ' + column.name + ': ' + duplicate)
  })
}

// Parse the CSV rows
function parseData (data, schema) {
  // Load and validate columns
  var columns = mapHeadersToColumns(data.headers, schema)
  validateColumns(columns, schema)

  // Load and validate rows
  var mapRow = _.partial(parseRow, _, columns)
  var rows = _.map(data.rows, mapRow)
  validateRows(rows, schema)
  return rows
}

// Parse CSV data from a stream
function parseStream (stream, schema) {
  return loadStream(stream).then(function (data) {
    return parseData(data, schema)
  })
}

// Load CSV data from a stream
function loadStream (stream) {
  return new Promise(function (resolve, reject) {
    var headers
    var rows = []

    csv.fromStream(stream, {trim: true, ignoreEmpty: true})
      .on('data', function (data) {
        if (!headers) headers = data
        else rows.push(data)
      })
      .on('end', function () {
        resolve({headers: headers, rows: rows})
      })
  })
}

module.exports = {
  Error: CsvError,
  loadStream: loadStream,
  parseData: parseData,
  parseStream: parseStream
}
