const _ = require('underscore')
const ActiveKnex = require('active-knex')

module.exports = function (knex, Rbac) {
  var Email = ActiveKnex.Schema.create(knex, {
    tableName: 'emails',
    keys: ['email'],

    fields: {
      createdAt: true
    },

    queries: {
      // Where the person meets this search string
      whereSearch: function (search) {
        if (!search || !search.length) return this
        var like = '%' + search + '%'
        return this.whereRaw('emails.email LIKE ?', like)
      },

      // Helpers for filtering by query
      whereQuery: function (query) {
        // Parse parameters
        var order = query.order
        var lastId = parseInt(query.last, 10)
        var limit = parseInt(query.limit, 10)
        var search = query.search

        // Execute query
        var qb = this
        if (limit) qb = qb.limit(limit)
        if (search) qb = qb.whereSearch(search)
        if (order) qb = qb.afterId(order, lastId)
        return qb
      }
    },

    relations: {
      people: function () {
        return ActiveKnex.Relation.HasMany({
          model: Rbac.Person,
          foreignKey: 'email_id'
        })
      }
    }
  })

  // // Ensure the email is lowercase
  // Email.before('save', function(row) {
  //   var rows = ActiveKnex.util.arrayify(row);
  //   _.each(rows, function(row) {
  //     if (row.email) row.email = Email.format(row.email);
  //   });
  // });

  // Format an email to the database
  Email.format = function (email) {
    return email.toLowerCase()
  }

  // Find row by email
  Email.findByEmail = function (email, trx) {
    return Email.transacting(trx).where('email', email).first()
  }

  // Find id by email
  Email.findId = function (email, trx) {
    return Email.findByEmail(email, trx).then(function (row) {
      if (row) return row.id
    })
  }

  // Find or create row by email
  Email.findOrCreateByEmail = function (email, trx) {
    return Email.findByEmail(email, trx).then(function (row) {
      return row || Email.create({email: email}, trx)
    })
  }

  // Find or create by email
  Email.findOrCreateId = function (email, trx) {
    return Email.findOrCreateByEmail(email, trx).then(function (row) {
      if (row) return row.id
    })
  }

  // Bulk create email records
  // Returns: returns email records sorted by email
  // NOTE: this function is extremely defensive
  Email.bulkFindOrCreateByEmail = function (emails, trx) {
    // Format all the emails
    emails = _.chain(emails).compact().map(Email.format).uniq().value()

    // Find the existings rows with these email emailes
    return Email.transacting(trx).whereIn('email', emails).orderBy('email').then(function (existingRows) {
      // Determine the missing emails
      var existingEmails = _.pluck(existingRows, 'email')
      var missingEmails = _.difference(emails, existingEmails)

      // Create missing emails
      var missing = _.map(missingEmails, function (email) { return {email: email} })
      return Email.create(missing, trx).then(function (newRows) {
        // Merge the two together
        var union = _.union(existingRows, newRows)
        return _.sortBy(union, 'email')
      })
    })
  }

  // Load email onto the objects (via emailId)
  // object: object or an array of objects
  Email.loadEmail = function (object, trx) {
    // Create a sorted list of objects
    var objects = ActiveKnex.util.arrayify(object)
    objects = _.chain(objects).filter(function (object) {
      return object.emailId
    }).sortBy('emailId').value()

    // Load the emails
    var emailIds = _.pluck(objects, 'emailId')
    return Email.transacting(trx).whereIn('id', emailIds).orderBy('id').then(function (rows) {
      var rowIndex = 0
      var row = rows[rowIndex]
      _.each(objects, function (object) {
        // Increment the row
        while (row && object.emailId > row.id) {
          rowIndex += 1
          row = rows[rowIndex]
        }

        // Match? Email.
        if (row && object.emailId === row.id) object.email = row.email
      })
    }).return(object)
  }

  // Load emailId onto the objects (via email, create if necessary)
  // object: object or an array of objects
  Email.loadEmailId = function (object, trx) {
    // Create a sorted list of objects
    var objects = ActiveKnex.util.arrayify(object)
    objects = _.chain(objects).filter(function (object) {
      return object.email
    }).map(function (object) {
      object.email = Email.format(object.email)
      return object
    }).sortBy('email').value()

    // Load the emails
    var emails = _.pluck(objects, 'email')
    return Email.bulkFindOrCreateByEmail(emails, trx).then(function (rows) {
      var rowIndex = 0
      var row = rows[rowIndex]
      _.each(objects, function (object) {
        // Increment the row
        while (row && object.email > row.email) {
          rowIndex += 1
          row = rows[rowIndex]
        }

        // Match? Email.
        if (row && object.email === row.email) object.emailId = row.id
      })
    }).return(object)
  }

  return Email
}
