const _ = require('underscore')
const ActiveKnex = require('active-knex')
const {pluckId} = require('../util')

module.exports = function (knex, Rbac) {
  var Group = ActiveKnex.Schema.create(knex, {
    tableName: 'groups',
    keys: ['organizationId', 'name'],

    fields: {
      archived: { type: 'boolean' },
      createdAt: true,
      updatedAt: true
    },

    joins: {
      group_members: function (qb) {
        return qb.leftJoin('group_members', 'groups.id', 'group_members.group_id')
      },

      organizations: function (qb) {
        return qb.leftJoin('organizations', 'groups.organization_id', 'organizations.id')
      }
    },

    queries: {
      // Where the person has not been archived
      whereActive: function () {
        return this.where('groups.archived', false)
      },

      // Groups within this organization
      whereOrganizationId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereIn('groups.organization_id', ids)
      },

      // Groups this person is a member of
      wherePersonId: function (id) {
        return this.whereRaw(`groups.id IN (SELECT group_id FROM group_members WHERE group_members.person_id = ?)`, id)
      },

      // Groups this user is allowed to see
      whereUserCan: function (user) {
        if (!user) return this.whereRaw('FALSE')
        if (user.admin) return this
        return this.whereUserIn(user)
      },

      // Organizations this user belongs to
      whereUserIn: function (user) {
        var group = 'groups.id IN (SELECT group_id FROM group_members INNER JOIN people ON people.id = group_members.person_id WHERE people.email_id IN (?))'
        var organization = 'groups.organization_id IN (SELECT organization_id FROM people WHERE email_id IN (?))'
        var sql = '( ' + group + ' OR ' + organization + ')'
        return this.whereRaw(sql, [user.emailId, user.emailId])
      },

      // Find groups with name that match the search text
      whereSearch: function (search) {
        var like = '%' + search + '%'
        return this.whereRaw('(groups.name LIKE ?)', [like])
      },

      // Find groups where the group has no members
      whereEmpty: function () {
        return this.whereRaw('(SELECT COUNT(person_id) FROM group_members WHERE group_id = groups.id) = 0')
      },

      // Find groups that match the search query
      whereQuery: function (query) {
        // Parse parameters
        var order = query.order || 'id'
        var limit = parseInt(query.limit, 10)
        var include = ActiveKnex.util.arrayify(query.include)
        var exclude = ActiveKnex.util.arrayify(query.exclude)
        var organizationId = parseInt(query.organization, 10)
        var search = query.search
        if (query.archived !== undefined) var archived = query.archived === 'true'

        // Execute query
        var qb = this
        if (limit) qb = qb.limit(limit)
        if (include.length) qb = qb.whereIn('groups.id', include)
        if (exclude.length) qb = qb.whereNotIn('groups.id', exclude)
        if (organizationId) qb = qb.whereOrganizationId(organizationId)
        if (search) qb = qb.whereSearch(search)
        if (!archived) qb = qb.whereActive()
        return qb.after(order, query.after)
      },

      // Count the number of active people in the group
      countPeople: function () {
        var count = knex.raw('COUNT(CASE WHEN people.archived <> 1 THEN 1 END) AS people_count')
        return this.joins('group_members.people').groupBy('groups.id').select(count)
      }
    },

    relations: {
      members: function () {
        return ActiveKnex.Relation.HasMany({
          model: Group.Member,
          foreignKey: 'group_id'
        })
      }
    }
  })

  // Attach member table
  Group.Member = require('./group/member')(knex, Rbac)

  // Find or create a group by name
  Group.findByName = function (organization, name, trx) {
    var organizationId = pluckId(organization)
    return Group.transacting(trx).where({
      organizationId: organizationId,
      name: name
    }).first()
  }

  // Find or create a group by name
  Group.findOrCreateByName = function (organization, name, trx) {
    var organizationId = pluckId(organization)
    return Group.findOrCreate({
      organizationId: organizationId,
      name: name
    }, trx).then(function (group) {
      return !group.archived ? group : Group.update(group, {archived: false}, trx)
    })
  }

  // Determine if these groups have the same name
  Group.isNameEqual = function (group, name) {
    return group.name.toLowerCase() === name.toLowerCase()
  }

  // Helper to filter groups by organization id
  Group.whereOrganizationId = function (organizationId, trx) {
    return Group.transacting(trx).whereOrganizationId(organizationId)
  }

  // Helper to filter groups for this person
  Group.wherePersonId = function (personId, trx) {
    return Group.transacting(trx).wherePersonId(personId)
  }

  // Remove this group/groups from sharing
  Group.removeFromPermissions = function (group) {
    var groups = ActiveKnex.util.arrayify(group)
    var organizationId = _.first(groups).organizationId
    var variants = _.map(groups, Rbac.Variant.fromGroup)
    var variables = _.invoke(variants, 'toString')
    return Rbac.Permission.removeVariable(variables, organizationId)
  }

  // Archive this group and remove all references to it
  Group.archive = function (group) {
    var groups = ActiveKnex.util.arrayify(group)
    var groupIds = _.map(groups, pluckId)
    return Group.removeFromPermissions(group).then(function () {
      return Group.query().whereIn('id', groupIds).update({archived: true})
    })
  }

  // Remove this group from existance
  Group.purge = function (group) {
    var groups = ActiveKnex.util.arrayify(group)
    var groupIds = _.map(groups, pluckId)
    return Group.removeFromPermissions(group).then(function () {
      return Group.query().whereIn('id', groupIds).del()
    })
  }

  // Ensure groups exist for all these names on the organization
  Group.bulkCreate = function (organization, name, trx) {
    var names = ActiveKnex.util.arrayify(name)
    var organizationId = pluckId(organization)
    return Group.transacting(trx).whereOrganizationId(organizationId).whereIn('name', names).then(function (groups) {
      // Find any archived groups
      var archived = _.where(groups, {archived: true})
      if (!archived.length) return groups

      // Unarchive archived groups
      var archivedIds = _.pluck(archived, 'id')
      return Group.transacting(trx).whereIn('id', archivedIds).update({archived: false}).then(function () {
        _.each(archived, function (group) { group.archived = false })
      }).return(groups)
    }).then(function (groups) {
      // Determine missing names
      var missingNames = _.reject(names, function (name) {
        return _.any(groups, function (group) {
          return Group.isNameEqual(group, name)
        })
      })

      // Create missing groups
      var missing = _.map(missingNames, function (name) {
        return {organizationId: organization.id, name: name}
      })

      return Group.create(missing, trx).then(function (newGroups) {
        // Merge the two together
        var union = _.union(groups, newGroups)
        return _.sortBy(union, 'name')
      })
    })
  }

  return Group
}
