const Promise = require('bluebird')
const _ = require('underscore')
const ActiveKnex = require('active-knex')
const {pluckId} = require('../../util')

module.exports = function (knex, Rbac) {
  var Member = ActiveKnex.Schema.create(knex, {
    tableName: 'group_members',
    keys: ['groupId', 'personId'],

    fields: {
      createdAt: true
    },

    joins: {
      groups: function (qb) {
        return qb.leftJoin('groups', 'group_members.group_id', 'groups.id')
      },

      people: function (qb) {
        return qb.leftJoin('people', 'group_members.person_id', 'people.id')
      }
    },

    queries: {
      whereGroupId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereIn('group_members.group_id', ids)
      },

      wherePersonId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereIn('group_members.person_id', ids)
      }
    }
  })

  _.extend(Member, {
    whereGroupId: function (groupId) {
      return Member.query().whereGroupId(groupId)
    },

    wherePersonId: function (personId) {
      return Member.query().wherePersonId(personId)
    },

    // Remove people from this group
    removeGroup: function (group) {
      var groups = ActiveKnex.util.arrayify(group)
      var groupIds = _.map(groups, pluckId)
      return Member.query().whereIn('group_id', groupIds).delete()
    },

    // Remove these person/people from this group
    removeFromGroup: function (group, person, trx) {
      var groupId = pluckId(group)
      var people = ActiveKnex.util.arrayify(person)
      var personIds = _.map(people, pluckId)
      return Member.transacting(trx).where('group_id', groupId).whereIn('person_id', personIds).delete()
    },

    // Remove these person/people from the groups within the organization
    // organization: organization to remove members from
    removeFromOrganization: function (organization) {
      var organizationId = pluckId(organization)
      return Member.query().whereRaw('group_id IN (SELECT id FROM groups WHERE organization_id IN (?))', organizationId).delete()
    },

    // Bulk add people to a group
    bulkAdd: function (group, person, trx) {
      var groupId = pluckId(group)
      var people = ActiveKnex.util.arrayify(person)
      var personIds = _.map(people, pluckId)
      return Member.transacting(trx).where('group_id', groupId).whereIn('person_id', personIds).then(function (members) {
        var existing = _.pluck(members, 'personId')
        var missing = _.chain(personIds).difference(existing).map(function (personId) {
          return {groupId: group.id, personId: personId}
        }).value()
        if (missing.length) return Member.create(missing, trx)
      })
    },

    // Bulk create people onto a group
    bulkCreate: function (group, entry, trx) {
      var entries = ActiveKnex.util.arrayify(entry)
      return Rbac.Person.bulkCreate(group.organizationId, entries, trx).then(function (people) {
        return Member.bulkAdd(group, people, trx).return(people)
      })
    },

    // Bulk assign an array of personId with their groupIds to the organization
    bulkAssign: function (organization, assignments, trx) {
      var personIds = _.pluck(assignments, 'personId')
      var organizationId = pluckId(organization)
      return Member.transacting(trx).joins('groups').where('organization_id', organizationId).whereIn('person_id', personIds).select('group_members.*').then(function (members) {
        // Determine what to remove
        var removeIds = _.chain(members).reject(function (member) {
          var assignment = _.findWhere(assignments, {personId: member.personId})
          return _.contains(assignment.groupIds, member.groupId)
        }).pluck('id').value()

        // Determine what to add
        var newMembers = _.chain(assignments).map(function (assignment) {
          var personId = assignment.personId
          return _.chain(assignment.groupIds).reject(function (groupId) {
            return _.findWhere(members, {personId: personId, groupId: groupId})
          }).map(function (groupId) {
            return {groupId: groupId, personId: personId}
          }).value()
        }).flatten().value()

        return Promise.all([
          Member.transacting(trx).whereIn('id', removeIds).delete(),
          Member.create(newMembers, trx)
        ])
      })
    }
  })

  return Member
}
