const _ = require('underscore')
const ActiveKnex = require('active-knex')

module.exports = function (knex, Rbac) {
  var Organization = ActiveKnex.Schema.create(knex, {
    tableName: 'organizations',

    fields: {
      createdAt: true,
      updatedAt: true
    },

    joins: {
      people: function (qb) {
        return qb.leftJoin('people', 'organizations.id', 'people.organization_id')
      }
    },

    queries: {
      // Organizations this user is allowed to see
      whereUserCan: function (user) {
        if (!user) return this.whereRaw('FALSE')
        if (user.admin) return this
        return this.whereUserIn(user)
      },

      // Organizations this user belongs to
      whereUserIn: function (user) {
        return this.whereRaw('id IN (SELECT organization_id FROM people WHERE email_id = ? AND archived = false)', user.emailId)
      },

      // Organizations that have this domain
      whereDomain: function (domain) {
        return this.where('organizations.domain', domain)
      },

      // Organizations that match this search criteria
      whereSearch: function (search) {
        var like = '%' + search + '%'
        return this.whereRaw('(organizations.name LIKE ?)', [like])
      }
    },

    relations: {
      people: function () {
        return ActiveKnex.Relation.HasMany({
          model: Rbac.Person,
          foreignKey: 'organization_id'
        })
      },

      groups: function () {
        return ActiveKnex.Relation.HasMany({
          model: Rbac.Group,
          foreignKey: 'organization_id'
        })
      },

      permissions: function () {
        return ActiveKnex.Relation.HasMany({
          model: Rbac.Permission,
          foreignKey: 'organization_id'
        })
      }
    }
  })

  _.extend(Organization, {
    createBy: function (params) {
      var owner = params.owner
      var organization = _.omit(params, 'owner')
      return Organization.create(organization).then(function (organization) { // Create organization
        return Rbac.Group.create({ // Create Admin group
          organizationId: organization.id,
          name: 'Admin'
        }).then(function (group) {
          return Rbac.Person.findOrCreateBy({ // Create Person
            organizationId: organization.id,
            email: owner.email,
            fullName: owner.fullName,
            member: true
          }).then(function (person) {
            return Rbac.Group.Member.create({ // Add Person to Admin Group
              groupId: group.id,
              personId: person.id
            }).then(function () {
              return Rbac.Privilege.query()
            }).then(function (privileges) { // Assign Admin group to all privilegs
              var permissions = _.map(privileges, function (privilege) {
                return {
                  privilegeId: privilege.id,
                  organizationId: organization.id,
                  expression: 'G' + group.id
                }
              })
              return Rbac.Permission.create(permissions)
            })
          })
        }).return(organization)
      })
    },

    // Organizations this user is allowed to see
    whereUserCan: function (user) {
      return Organization.query().whereUserCan(user)
    },

    // Organizations this user belongs to
    whereUserIn: function (user) {
      return Organization.query().whereUserIn(user)
    },

    findByDomain: function (domain) {
      return Organization.query().whereDomain(domain).first()
    },

    loadGroups: function (organization, trx) {
      var qb = Rbac.Group.whereOrganizationId(organization.id, trx).whereActive()
      return qb.select('groups.*').countPeople().then(function (groups) {
        organization.groups = groups
      }).return(organization)
    },

    // Get the people counts for the organization
    getPeopleCounts: function (organization, trx) {
      return Rbac.Person.whereOrganizationId(organization.id, trx).countPeople().first()
    },

    // Load the people counts onto the organization
    loadPeopleCounts: function (organization, trx) {
      return Organization.getPeopleCounts(organization, trx).then(function (counts) {
        return _.extend(organization, counts)
      })
    },

    // Load the privileges onto the organization for this user
    loadPrivileges: function (organization, user) {
      return Rbac.Permission.getPrivileges(user, {
        organizationId: organization.id,
        resourceType: null
      }).then(function (privileges) {
        organization.privileges = privileges
      }).return(organization)
    },

    // Determine if one of the loadPrivileges has been loaded
    can: function (organization, privilege) {
      return _.includes(organization.privileges, privilege)
    }
  })

  return Organization
}
