const Promise = require('bluebird')
const _ = require('underscore')
const ActiveKnex = require('active-knex')

module.exports = function (knex, Rbac) {
  var Permission = Rbac.Schema.create(knex, {
    tableName: 'permissions',
    keys: ['privilegeId', 'organizationId', 'resourceType', 'resourceId'],

    fields: {
      createdAt: true,
      updatedAt: true
    },

    joins: {
      privileges: function (qb) {
        return qb.leftJoin('privileges', 'permissions.privilege_id', 'privileges.id')
      }
    },

    queries: {
      // Helper to join the email table
      joinsPrivileges: function () {
        return this.joins('privileges').select('permissions.*', 'privileges.name as privilege')
      },

      // Permissions this user is allowed to see
      whereUser: function (user) {
        return this.whereRaw('organization_id IN (SELECT organization_id FROM people WHERE email_id = ? AND archived = false)', user.emailId)
      },

      wherePrivilege: function (name) {
        var names = ActiveKnex.util.arrayify(name)
        return this.joins('privileges').whereIn('privileges.name', names)
      },

      whereOrganizationId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereIn('organization_id', ids)
      },

      whereResourceType: function (type) {
        return this.where('resource_type', type)
      },

      whereResourceId: function (id, allowNull) {
        if (!allowNull) return this.where('resource_id', id)
        return this.whereRaw('(resource_id = ? OR resource_id IS NULL)', id)
      },

      whereScope: function (scope) {
        if (scope.privilege) this.wherePrivilege(scope.privilege)

        if (scope.organizationId) this.whereOrganizationId(scope.organizationId)

        if (scope.resourceType) this.whereResourceType(scope.resourceType)
        else if (scope.resourceType === null) this.whereNull('resource_type')

        if (scope.resourceId) this.whereResourceId(scope.resourceId, true)
        else if (scope.resourceId === null) this.whereNull('resource_type')

        return this
      }
    },

    relations: {
      privilege: function () {
        return ActiveKnex.Relation.BelongsTo({
          model: Rbac.Privilege,
          foreignKey: 'privilege_id'
        })
      }
    }
  })

  // Helper to get the permissions on this organization
  Permission.whereOrganizationId = function (organizationId, trx) {
    return Permission.transacting(trx).whereOrganizationId(organizationId)
  }

  // Evaluate the permissions for the user
  Permission.filterByUser = function (permissions, user) {
    if (!permissions.length) return Promise.resolve(permissions)

    var sql = _.map(permissions, function (permission) {
      var expressionSql = Rbac.Person.expressionToSql(permission.expression)
      return `SELECT ${permission.id} AS permission_id FROM people WHERE ` + expressionSql + ` AND people.organization_id = ${permission.organizationId} AND people.email_id = ${user.emailId} AND people.archived = FALSE`
    }).join(' UNION ')

    return knex.raw(sql).spread(function (rows) {
      var permissionIds = _.pluck(rows, 'permission_id')
      return _.filter(permissions, function (permission) {
        return _.contains(permissionIds, permission.id)
      })
    })
  }

  // Get the permissions this user can access
  Permission.whereUser = function (user, scope) {
    var qb = Permission.query().whereUser(user).whereScope(scope)
    return qb.joinsPrivileges().then(function (permissions) {
      return Permission.filterByUser(permissions, user)
    })
  }

  // Get the privileges the user has on this scope
  Permission.getPrivileges = function (user, scope) {
    return Permission.whereUser(user, scope).then(function (permissions) {
      return _.chain(permissions).pluck('privilege').uniq().value()
    })
  }

  // Create Permissions by privilege name, looks up the privilege id
  Permission.createBy = function (entry, trx) {
    var entries = ActiveKnex.util.arrayify(entry)
    var names = _.pluck(entries, 'privilege')
    return Rbac.Privilege.bulkCreate(names, trx).then(function (privileges) {
      var permissions = _.map(entries, function (entry) {
        var privilege = _.findWhere(privileges, {name: entry.privilege})
        var permission = _.omit(entry, 'privilege')
        permission.privilegeId = privilege.id
        return permission
      })
      return Permission.create(permissions, trx)
    })
  }

  // Remove this variable(s) from the expressions on this organization
  Permission.removeVariable = function (variables, organizationId, trx) {
    return Permission.whereOrganizationId(organizationId, trx).then(function (permissions) {
      return Promise.each(permissions, function (permission) {
        if (permission.expression) var expression = Rbac.Expression.removeVariable(permission.expression, variables)
        if (expression !== permission.expression) return Permission.update(permission, {expression: expression}, trx)
      })
    })
  }

  return Permission
}
