const _ = require('underscore')
const util = require('util')
const Promise = require('bluebird')
const ActiveKnex = require('active-knex')
const {pluckId} = require('../util')
const csv = require('../csv')

module.exports = function (knex, Rbac) {
  var Person = ActiveKnex.Schema.create(knex, {
    tableName: 'people',
    keys: ['organizationId', 'emailId'],

    fields: {
      member: { type: 'boolean' },
      archived: { type: 'boolean' },
      createdAt: true,
      updatedAt: true
    },

    joins: {
      group_members: function (qb) {
        return qb.leftJoin('group_members', 'people.id', 'group_members.person_id')
      },

      organizations: function (qb) {
        return qb.leftJoin('organizations', 'people.organization_id', 'organizations.id')
      },

      emails: function (qb) {
        return qb.leftJoin('emails', 'people.email_id', 'emails.id')
      }
    },

    queries: {
      // Helper to join the email table
      joinEmail: function () {
        return this.joins('emails').select('people.*', 'emails.email')
      },

      // Where the person has not been archived
      whereActive: function () {
        return this.where('people.archived', false)
      },

      // Where the person is a member of the organization
      whereOrganizationId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereIn('people.organization_id', ids)
      },

      // Where the person is a member of the group
      whereGroupId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereRaw('(SELECT COUNT(*) FROM group_members WHERE person_id = people.id AND group_id IN (?)) > 0', ids)
      },

      // Where the person is NOT a member of the group
      whereNotGroupId: function (id) {
        var ids = ActiveKnex.util.arrayify(id)
        return this.whereRaw('(SELECT COUNT(*) FROM group_members WHERE person_id = people.id AND group_id IN (?)) = 0', ids)
      },

      // Where the person has this email id
      whereEmailId: function (emailId) {
        var emailIds = ActiveKnex.util.arrayify(emailId)
        return this.whereIn('email_id', emailIds)
      },

      // Where the person has this email address
      whereEmail: function (email) {
        var emails = ActiveKnex.util.arrayify(email)
        return this.whereRaw('people.email_id IN (SELECT id FROM emails WHERE emails.email IN (?))', emails)
      },

      // Where the person meets this search string
      whereSearch: function (search) {
        if (!search || !search.length) return this
        var like = '%' + search + '%'
        return this.joins('emails').whereRaw('(emails.email LIKE ? OR people.full_name LIKE ? OR people.employee_id LIKE ?)', [like, like, like])
      },

      // People this user is allowed to see
      whereUserCan: function (user) {
        if (!user) return this.whereRaw('FALSE')
        if (user.admin) return this
        return this.whereRaw('people.organization_id IN (SELECT organization_id FROM people WHERE email_id = ?)', user.emailId)
      },

      whereExpression: function (expression) {
        var sqlExpression = Person.expressionToSql(expression)
        return this.whereRaw(sqlExpression)
      },

      whereQuery: function (query) {
        // Parse parameters
        var expression = query.expression
        var search = query.search
        var order = query.order
        var lastId = parseInt(query.last, 10)
        var limit = parseInt(query.limit, 10)
        var include = ActiveKnex.util.arrayify(query.include)
        var exclude = ActiveKnex.util.arrayify(query.exclude)
        var groupId = parseInt(query.group, 10)
        var organizationId = parseInt(query.organization, 10)
        if (query.member !== undefined) var member = query.member === 'true'
        if (query.archived !== undefined) var archived = query.archived === 'true'

        // Execute query
        var qb = this
        if (expression) qb = qb.whereExpression(expression)
        if (search) qb = qb.whereSearch(search)
        if (limit) qb = qb.limit(limit)
        if (include.length) qb = qb.whereIn('people.id', include)
        if (exclude.length) qb = qb.whereNotIn('people.id', exclude)
        if (groupId && groupId > 0) qb = qb.whereGroupId(groupId)
        if (groupId && groupId < 0) qb = qb.whereNotGroupId(-groupId)
        if (organizationId) qb = qb.whereOrganizationId(organizationId)
        if (member !== undefined) qb = qb.where('people.member', member)
        if (!archived) qb = qb.whereActive()
        if (order) qb = qb.afterId(order, lastId)
        return qb
      },

      // Count the people in the query
      countPeople: function () {
        var peopleCount = knex.raw('COUNT(CASE WHEN !people.archived THEN 1 END) AS people_count')
        var membersCount = knex.raw('COUNT(CASE WHEN people.member AND !people.archived THEN 1 END) AS members_count')
        var friendsCount = knex.raw('COUNT(CASE WHEN !people.member AND !people.archived THEN 1 END) AS friends_count')
        return this.select(peopleCount).select(membersCount).select(friendsCount)
      }
    },

    relations: {
      organization: function () {
        return ActiveKnex.Relation.BelongsTo({
          model: Rbac.Organization,
          foreignKey: 'organization_id'
        })
      }
    },

    orders: {
      'email': 'emails.email',
      '-email': '-emails.email'
    }
  })

  // Schema for CSV parsing
  Person.csvSchema = [{
    name: 'ID',
    field: 'employeeId',
    regEx: /id/i,
    unique: true
  }, {
    name: 'Email',
    field: 'email',
    regEx: /email/i,
    type: 'email',
    required: true,
    unique: true
  }, {
    name: 'Name',
    field: 'fullName',
    regEx: /name/i,
    required: true
  }, {
    name: 'Groups',
    field: 'groups',
    regEx: /groups?/i,
    type: 'array'
  }]

  // Find or Create a person from these parameters
  Person.findOrCreateBy = function (params, trx) {
    return Rbac.Email.loadEmailId(params, trx).then(function (params) {
      var userParams = _.pick(params, 'email')
      params = _.omit(params, 'email')
      return Person.findOrCreate(params, trx).then(function (person) {
        return _.extend(person, userParams)
      })
    })
  }

  // Update or create a person from these parameters
  Person.updateOrCreateBy = function (params, trx) {
    return Rbac.Email.loadEmailId(params, trx).then(function (params) {
      var userParams = _.pick(params, 'email')
      params = _.omit(params, 'email')
      return Person.updateOrCreate(params, trx).then(function (person) {
        return _.extend(person, userParams)
      })
    })
  }

  // Find by user (order so active person beats archived person)
  Person.findByUser = function (user, organization, trx) {
    return Person.findByEmailId(user.emailId, organization, trx)
  }

  // Find by email id (order so active person beats archived person)
  Person.findByEmailId = function (emailId, organization, trx) {
    return Person.whereEmailId(emailId, organization, trx).first()
  }

  // Find by email (order so active person beats archived person)
  Person.findByEmail = function (email, organization, trx) {
    return Rbac.Email.findId(email, trx).then(function (emailId) {
      if (emailId) return Person.findByEmailId(emailId, organization, trx)
    })
  }

  // Filter people by organization
  Person.whereOrganizationId = function (organizationId, trx) {
    return Person.transacting(trx).whereOrganizationId(organizationId)
  }

  // Restrict to people with this emailId id (order so active person beats archived person)
  Person.whereEmailId = function (emailId, organization, trx) {
    var qb = Person.transacting(trx).whereEmailId(emailId)
    if (organization) qb = qb.whereOrganizationId(organization.id)
    return qb.joinEmail().orderBy(['archived', '-member'])
  }

  // Active people that match this expression on an organization. We don't want this to bleed into other organizations.
  Person.whereExpression = function (expression, organizationId, trx) {
    var qb = Person.transacting(trx)
    qb = qb.whereExpression(expression)
    qb = qb.whereOrganizationId(organizationId)
    qb = qb.whereActive()
    return qb.select('people.*')
  }

  // Load the group Ids onto these people
  Person.loadGroupIds = function (person) {
    var people = ActiveKnex.util.arrayify(person)
    var personIds = _.pluck(people, 'id')
    return Rbac.Group.Member.wherePersonId(personIds).then(function (members) {
      _.each(people, function (person) {
        person.groupIds = _.chain(members).where({personId: person.id}).pluck('groupId').value()
      })
    }).return(person)
  }

  // Update the person by these params
  Person.updateBy = function (person, changes) {
    return Rbac.Email.loadEmailId(changes).then(function () {
      var userParams = _.pick(changes, 'email')
      changes = _.omit(changes, 'email')
      return Person.update(person, changes).then(function () {
        return _.extend(person, userParams)
      })
    })
  }

  // Remove this person/people from permissions
  Person.removeFromPermissions = function (person, trx) {
    var people = ActiveKnex.util.arrayify(person)
    var organizationId = _.first(people).organizationId
    var variants = _.map(people, Rbac.Variant.fromPerson)
    var variables = _.invoke(variants, 'toString')
    return Rbac.Permission.removeVariable(variables, organizationId, trx)
  }

  // Archive this person/people for historical purposes
  Person.archive = function (person, trx) {
    var now = new Date()
    var people = ActiveKnex.util.arrayify(person)
    var personIds = _.map(people, pluckId)
    return Person.removeFromPermissions(people, trx).then(function () {
      return Person.transacting(trx).whereIn('id', personIds).update({archived: true, updatedAt: now})
    }).then(function () {
      _.each(people, (person) => { person.archived = true })
    }).return(person)
  }

  // Unrchive this person/people
  Person.unarchive = function (person, trx) {
    var now = new Date()
    var people = ActiveKnex.util.arrayify(person)
    var personIds = _.map(people, pluckId)
    return Person.transacting(trx).whereIn('id', personIds).update({archived: false, updatedAt: now}).then(function () {
      _.each(people, (person) => { person.archived = false })
    }).return(person)
  }

  // Purge this person/people from the DB forever
  Person.purge = function (person, trx) {
    var people = ActiveKnex.util.arrayify(person)
    var personIds = _.map(people, pluckId)
    return Person.removeFromPermissions(person, trx).then(function () {
      return Person.query().whereIn('id', personIds).del()
    })
  }

  // Bulk update/insert people to the organization
  Person.bulkUpsert = function (organization, entry, trx) {
    // Load email onto entries
    var entries = ActiveKnex.util.arrayify(entry)
    var organizationId = pluckId(organization)
    return Rbac.Email.loadEmailId(entries, trx).then(function () {
      return Person.transacting(trx).where('organization_id', organizationId)
    }).then(function (people) {
      // Create a new person
      function createPerson (params) {
        params.organizationId = organizationId
        return Person.bufferCreate(params, trx)
      }

      // Update person
      function updatePerson (person, params) {
        var keys = _.keys(params)
        var current = _.pick(person, keys)
        var changed = !_.isEqual(params, current)
        return changed ? Person.update(person, params, trx) : Promise.resolve(person)
      }

      // Reduce to update or create
      return Promise.map(entries, function (entry) {
        // Parameters
        var userParams = _.pick(entry, 'email')
        var params = _.omit(entry, 'email')

        // Find an existing person
        var person = _.find(people, function (person) { // By employee ID first
          return entry.employeeId && (person.employeeId === entry.employeeId)
        }) || _.find(people, function (person) { // By Email ID second
            return person.emailId === entry.emailId
          })

        // Create / Update person, then add user attributes back
        var promise = person ? updatePerson(person, params) : createPerson(params)
        return promise.then(function (person) {
          return _.extend(person, userParams)
        })
      })
    })
  }

  // Create Organization Members/Groups/Group Members
  // organization: Organization to add them to
  // entry: Object/Array of email,fullName and groups
  // trx: transaction to run in (optional)
  Person.bulkCreate = function (organization, entry, trx) {
    // Format the incoming data
    var entries = ActiveKnex.util.arrayify(entry)
    _.each(entries, function (entry) {
      entry.archived = false // Unarchive anyone archived
    })

    // Determine people params
    var peopleParams = _.map(entries, function (entry) {
      return _.omit(entry, 'groups')
    })

    // Create people
    return Person.bulkUpsert(organization, peopleParams, trx).then(function (people) {
      // Determine the group params
      var groupParams = _.map(entries, function (entry) {
        if (entry.groups) return _.chain(entry.groups).compact().uniq().value()
      })

      // Move on, if no groups
      var isGroups = _.compact(groupParams).length > 0
      if (!isGroups) return people

      // Assign to groups
      var assign = _.chain(people).map(function (person, index) {
        return {personId: person.id, groups: groupParams[index]}
      }).compact().value()
      return Person.bulkAssign(organization, assign, trx).return(people)
    })
  }

  // Bulk Assign people to groups, each person has a groups parameter as an array of names
  Person.bulkAssign = function (organization, entries, trx) {
    var names = _.chain(entries).pluck('groups').flatten().uniq().value()
    return Rbac.Group.bulkCreate(organization, names, trx).then(function (groups) { // Create groups
      var assignments = _.chain(entries).map(function (entry) {
        var assign = _.pick(entry, 'personId')
        assign.groupIds = _.chain(entry.groups).map(function (name) {
          return _.find(groups, function (group) {
            return Rbac.Group.isNameEqual(group, name)
          })
        }).pluck('id').compact().value()
        return assign
      }).compact().value()
      return Rbac.Group.Member.bulkAssign(organization, assignments, trx)
    })
  }

  // Custom Bulk Error
  Person.BulkError = function (message) {
    this.name = 'BulkError'
    this.message = (message || '')
  }
  util.inherits(Person.BulkError, Error)

  // Validate the entries coming in
  Person.bulkValidate = function (entries) {
    var fields = _.chain(Person.csvSchema).where({unique: true}).pluck('field').value()
    _.each(fields, function (field) {
      // Search for a duplicate
      var duplicate = _.chain(entries).pluck(field).compact().countBy().reduce(function (result, count, value) {
        return result || (count > 1 ? value : null)
      }, null).value()

      // If duplicate, throw error
      if (duplicate) throw new Person.BulkError('Duplicate ' + field + ': ' + duplicate)
    })
  }

  // Bulk load new members, while deleting members no longer on this list
  Person.bulkLoad = function (organization, entries, trx) {
    // Validate the incoming data
    Person.bulkValidate(entries)

    // Format the incoming data
    _.each(entries, function (entry) {
      entry.member = true // Everyone loaded is a member
    })

    // Create/Update the new members
    return Person.bulkCreate(organization, entries, trx).then(function (people) {
      var personIds = _.pluck(people, 'id')
      var organizationId = pluckId(organization)

      var qb = Person.query()
      qb = qb.where('organization_id', organizationId)
      qb = qb.whereNotIn('id', personIds)
      qb = qb.where('member', true)
      return qb.then(function (removing) { // Archive members not on the list
        if (removing.length) return Person.archive(removing, trx)
      }).return(people)
    })
  }

  // Parse the bulk upload stream (throw error if invalid)
  Person.bulkParse = function (stream, user) {
    return csv.parseStream(stream, this.csvSchema).then(function (entries) {
      // Must have entries
      if (!entries.length) throw new csv.Error('The file provided is empty')

      // Must have my email (if user specified and not an admin)
      if (user && !user.admin) {
        var email = Rbac.Email.format(user.email)
        var emails = _.chain(entries).pluck('email').compact().map(Rbac.Email.format).value()
        var valid = _.contains(emails, email)
        if (!valid) throw new csv.Error('Could not find your email address: ' + email)
      }

      return entries
    })
  }

  // Bulk upload this filestream to the organizations (throw error if invalid)
  // organization: Organization to upload members to
  // stream: filesteam with rows of emails, names, etc.
  // user: user uploading the csv (optional, for validation purposes)
  Person.bulkUpload = function (organization, stream, user) {
    return Person.bulkParse(stream, user).then(function (entries) {
      return Person.bulkLoad(organization, entries)
    })
  }

  // Merge entries from two data sources before bulk load
  Person.bulkMerge = function (entries) {
    entries = _.clone(entries)

    var index = 0
    while (index < entries.length) {
      var entry = entries[index]
      var nextIndex = index + 1
      while (nextIndex < entries.length) {
        var next = entries[nextIndex]
        if (entry.employeeId === next.employeeId && entry.email === next.email) {
          entry.groups = _.union(entry.groups, next.groups)
          entries.splice(nextIndex, 1)
        } else {
          nextIndex++
        }
      }
      index++
    }

    return entries
  }

  // Bulk download this organization to the filestream
  // organization: Organization to download from (groups/people should be loaded)
  Person.bulkDownload = function (organization, stream) {
    return Rbac.Organization.loadGroups(organization).then(function () {
      return Rbac.Person.query().joinEmail().where({organizationId: organization.id}).whereActive()
    }).then(function (people) {
      return Rbac.Person.loadGroupIds(people)
    }).then(function (people) {
      // Determine if we have an employee ID field
      var hasEmployeeId = _.any(people, function (person) {
        return person.employeeId
      })

      // Create CSV header line
      var headers = ['Email', 'Name', 'Groups']
      if (hasEmployeeId) headers.unshift('External ID')
      var line = headers.join(',')
      stream.push(line + '\n')

      // Determine any groups to exclude from CSV schema
      var groupsColumn = _.findWhere(Person.csvSchema, {field: 'groups'})
      var excludeGroups = groupsColumn.exclude ? groupsColumn.exclude : []

      // Add line per-person
      _.chain(people).filter(function (person) {
        return person.member // Only include members
      }).each(function (person) {
        // Determine the group names
        var groups = _.chain(person.groupIds).map(function (groupId) {
          return _.findWhere(organization.groups, {id: groupId})
        }).compact().pluck('name').difference(excludeGroups).value()

        // Determine the fields
        var fields = [person.email, person.fullName].concat(groups)
        if (hasEmployeeId) fields.unshift(person.employeeId)

        // Serialize into a line
        var line = _.map(fields, function (field) {
          return '"' + field + '"'
        }).join(',')
        stream.push(line + '\n')
      })

      // Complete response
      stream.push(null)
    })
  }

  // Convert the expression in Pxx/Gxxx/Oxxx form to SQL
  Person.expressionToSql = function (expression) {
    // No expression, no results
    if (!expression) return 'FALSE'

    expression = expression.replace(/P(\d+)/g, '(people.id = $1)') // People
    expression = expression.replace(/G(\d+)/g, '(people.id IN (SELECT person_id FROM group_members WHERE group_id = $1))') // Groups
    expression = expression.replace(/O(\d+)/g, '(people.organization_id = $1 AND people.member = true)') // Organizations
    return '(' + expression + ')'
  }

  // Evaluate the permissions on the people table
  Person.evaluatePermissions = function (permissions) {
    var qb = Person.query().select('people.id')

    _.each(permissions, function (permission) {
      var expressionSql = [
        Person.expressionToSql(permission.expression),
        'people.organization_id = ' + permission.organizationId,
        'people.archived = FALSE'
      ].join(' AND ')

      var selectSql = knex.raw(expressionSql + ' AS permission_' + permission.id)
      qb.select(selectSql)
    })

    return qb
  }

  // Variants -----------------------------------------------------------------

  // Find the variants for this person
  Person.getVariants = function (person, trx) {
    return Promise.all([
      Rbac.Organization.transacting(trx).joins('people').where('people.member', true).where('people.email_id', person.emailId).distinct('organizations.*'),
      Rbac.Group.transacting(trx).joins('group_members.people').where('people.email_id', person.emailId).distinct('groups.*'),
      Rbac.Person.transacting(trx).whereEmailId(person.emailId)
    ]).spread(function (organizations, groups, people) {
      return _.union(
        _.map(organizations, Rbac.Variant.fromOrganization),
        _.map(groups, Rbac.Variant.fromGroup),
        _.map(people, Rbac.Variant.fromPerson)
      )
    })
  }

  return Person
}
