const _ = require('underscore')
const url = require('url')
const moment = require('moment')
const Promise = require('bluebird')
const ActiveKnex = require('active-knex')
const crypto = require('crypto')

module.exports = function (knex, Rbac) {
  var User = ActiveKnex.Schema.create(knex, {
    tableName: 'users',
    keys: ['emailId'],

    fields: {
      admin: { type: 'boolean' },
      lastSignedInAt: {type: 'date'},
      resetSessionAt: {type: 'date'},
      authTokenSentAt: {type: 'date'},
      createdAt: true
    },

    joins: {
      emails: function (qb) {
        return qb.leftJoin('emails', 'users.email_id', 'emails.id')
      }
    },

    queries: {
      // Helper to join the email table
      joinEmail: function () {
        return this.joins('emails').select('users.*', 'emails.email')
      },

      // Where the person meets this search string
      whereSearch: function (search) {
        if (!search || !search.length) return this
        var like = '%' + search + '%'
        return this.joins('emails').whereRaw('emails.email LIKE ?', like)
      },

      // Helpers for filtering by query
      whereQuery: function (query) {
        // Parse parameters
        var order = query.order
        var lastId = parseInt(query.last, 10)
        var limit = parseInt(query.limit, 10)
        var search = query.search

        // Execute query
        var qb = this
        if (limit) qb = qb.limit(limit)
        if (search) qb = qb.whereSearch(search)
        if (order) qb = qb.afterId(order, lastId)
        return qb
      }
    },

    relations: {
      people: function () {
        return ActiveKnex.Relation.HasMany({
          model: Rbac.Person,
          key: 'email_id',
          foreignKey: 'email_id',
          query: function (ids, trx) {
            return Rbac.Person.transacting(trx).joinEmail().whereIn('people.email_id', ids)
          }
        })
      }
    }
  })

  // Auth Keys ----------------------------------------------------------------

  // Pick the auth keys off, return nothing if they aren't present
  User.pickAuthKeys = function (params) {
    var keys = _.pick(params, User.keys)
    if (_.keys(keys).length) return keys
  }

  // Load auth keys to incoming parameters
  User.loadAuthKeys = function (params, trx) {
    return Rbac.Email.loadEmailId(params, trx)
  }

  // Map incoming parameters to include auth keys, and strip the rest
  User.mapAuthKeys = function (params, trx) {
    return User.loadAuthKeys(params, trx).then(function () {
      return _.omit(params, 'email')
    })
  }

  // Load auth information
  User.loadAuth = function (user, trx) {
    return Rbac.Email.loadEmail(user, trx)
  }

  // Find / Update / Create ---------------------------------------------------

  // Find a user by id
  User.findById = function (id, trx) {
    return User.transacting(trx).where('users.id', id).joinEmail().first()
  }

  // Find a user
  User.findBy = function (params, trx) {
    return User.mapAuthKeys(params, trx).then(function (params) {
      var keys = User.pickAuthKeys(params)
      if (keys) return User.transacting(trx).where(keys).joinEmail().first()
    })
  }

  // Find or build a user
  User.findOrBuildBy = function (params, trx) {
    return User.mapAuthKeys(params, trx).then(function (params) {
      var keys = User.pickAuthKeys(params)
      if (keys) {
        return User.findBy(keys, trx).then(function (user) {
          return user || User.loadAuth(params, trx)
        })
      }
    })
  }

  // Find or create a user
  User.findOrCreateBy = function (params, trx) {
    return User.mapAuthKeys(params, trx).then(function (params) {
      var keys = User.pickAuthKeys(params)
      if (keys) {
        return User.findOrCreate(params, trx).then(function (user) {
          return User.loadAuth(user, trx)
        })
      }
    })
  }

  // Update or create a user
  User.updateOrCreateBy = function (params, trx) {
    return User.mapAuthKeys(params, trx).then(function (params) {
      var keys = User.pickAuthKeys(params)
      if (keys) {
        return User.updateOrCreate(params, trx).then(function (user) {
          return User.loadAuth(user, trx)
        })
      }
    })
  }

  // Set the parameters after a user signed in
  User.signIn = function (user) {
    return User.update(user, {
      lastSignedInAt: new Date(),
      authToken: null,
      authTokenSentAt: null
    })
  }

  // Authentication Token -----------------------------------------------------
  // An auth token allows the email to sign in via the URL

  // Generate a new auth token
  User.generateAuthToken = function () {
    var randomBytes = Promise.promisify(crypto.randomBytes, {context: crypto})
    return randomBytes(24).then(function (buf) {
      return buf.toString('hex')
    })
  }

  // Clear the authorization token data for this email
  User.clearAuthToken = function (email) {
    return Rbac.Email.update(email, {
      authToken: null,
      authTokenSentAt: null
    })
  }

  // Reset the authorization token data for this email
  User.resetAuthToken = function (email) {
    return User.generateAuthToken().then(function (token) {
      return User.update(email, {
        authToken: token,
        authTokenSentAt: new Date()
      })
    })
  }

  // Add the auth token for this email to the URL
  User.addAuthToken = function (email, uri) {
    var parsed = url.parse(uri, true)
    parsed.query.auth_token = email.authToken
    parsed.search = undefined
    return url.format(parsed)
  }

  // Find a email by auth token
  User.findByAuthToken = function (token) {
    return User.where('auth_token', token).joinEmail().first()
  }

  // Validate the auth token
  // sendAt: When the token was sent out (required)
  // minutes: Number of minutes to allow the token to be valid (optional: defaults to 15)
  User.isAuthTokenValid = function (sentAt, minutes) {
    minutes = minutes || 15
    var now = moment().toDate()
    var endsAt = moment(sentAt).add(minutes, 'minutes').toDate()
    return now >= sentAt && now <= endsAt
  }

  // Access Token -------------------------------------------------------------
  // An access token allows the server to access a URL on behalf
  // of a user, but does not log them in

  // Access token secret. Override if you see fit!
  User.accessTokenSecret = 'B61213564842D6642AF7A6C844DEED36B7CA86B4E8C6B1538A8B55FFF9D836E83A7C7C9436C8667B15B428C48F7361762FBE483F5F68B84A92C1'

  // Generate a access token
  User.generateAccessToken = function (userId, timestamp) {
    var value = [userId, User.accessTokenSecret, timestamp].join('')
    return crypto.createHash('sha1').update(value).digest('hex')
  }

  // Create the access parameters for this user
  User.generateAccessParams = function (user) {
    var timestamp = Math.floor(new Date().getTime() / 1000)
    return {
      access_id: user.id,
      access_timestamp: timestamp,
      access_token: User.generateAccessToken(user.id, timestamp)
    }
  }

  // Find a user by id, but the token and timestamp must match
  User.parseAccessParams = function (query, duration) {
    // Parse parameters, must have them all
    var userId = parseInt(query.access_id, 10)
    var timestamp = parseInt(query.access_timestamp, 10)
    var token = query.access_token
    if (!userId || !timestamp || !token) return

    // Check the token
    var expected = User.generateAccessToken(userId, timestamp)
    if (expected !== token) return

    // Check the time
    var now = Math.floor(new Date().getTime() / 1000)
    var delta = now - timestamp
    if (delta < duration) return userId
  }

  // Email --------------------------------------------------------------------

  // Find a user by an email address
  User.findByEmail = function (email, trx) {
    return User.findBy({email: email}, trx)
  }

  // Find or create a user by an email address
  User.findOrCreateByEmail = function (email, trx) {
    return User.findOrCreateBy({email: email}, trx)
  }

  // Reset Session ------------------------------------------------------------

  // Reset the sessions for users that match the search
  User.resetSessions = function (query) {
    query = query || {}

    var qb = User.query()
    if (query.id) qb.where('id', query.id)
    if (query.search) qb.whereSearch(query.search)
    return qb.update({reset_session_at: knex.raw('NOW()')})
  }

  return User
}
