/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const request = require('supertest')
const {expect} = require('chai')
const server = require('../server')
const Factory = require('../factory')
const {Group} = require('../rbac')

describe('Groups Controller', () => {
  var admin
  var user
  var member
  var friend
  var organization
  var group
  var otherGroup

  beforeEach(async () => {
    organization = await Factory.create('organization')
    member = await Factory.createPerson(organization, {member: true})
    friend = await Factory.createPerson(organization)

    group = await Factory.create('group', {organizationId: organization.id})
    await Factory.addGroupMember(group, friend)

    otherGroup = await Factory.create('group', {organizationId: organization.id})
    await Factory.create('group', {organizationId: 0})

    admin = await Factory.createUser({admin: true})
    user = await Factory.createUser()
  })

  describe('GET /groups', () => {
    it('should return 401', () => {
      return request(server)
        .get('/groups')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return groups (user)', () => {
      return request(server)
        .get('/groups?organization=' + organization.id)
        .set('From', member.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var groups = res.body
          expect(groups).to.have.length(2)
        })
    })

    it('should return groups (admin) @cleandb', () => {
      return request(server)
        .get('/groups')
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var groups = res.body
          expect(groups).to.have.length(3)
        })
    })
  })

  describe('POST /groups', () => {
    it('should return 401', () => {
      return request(server)
        .post('/organizations/' + organization.id + '/groups')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should NOT create a group on another organization', () => {
      return request(server)
        .post('/organizations/' + organization.id + '/groups')
        .send({organizationId: organization.id, name: 'New Group'})
        .set('From', user.email)
        .set('Accept', 'application/json')
        .expect(404)
    })

    it('should create a group', () => {
      return request(server)
        .post('/organizations/' + organization.id + '/groups')
        .send({name: 'New Group'})
        .set('From', member.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var group = res.body
          expect(group.organizationId).to.eql(organization.id)
          expect(group.name).to.eql('New Group')
        })
    })
  })

  describe('DELETE /groups/:id', () => {
    it('should return 401', () => {
      return request(server)
        .del('/groups/' + group.id)
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should delete the group', () => {
      return request(server)
        .del('/groups/' + group.id)
        .set('From', member.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          return Group.findById(group.id).then(function (group) {
            expect(group).to.be.undefined
          })
        })
    })
  })

  describe('POST /groups/:groupId/members', () => {
    it('should return 401', () => {
      return request(server)
        .post('/groups/' + group.id + '/members')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should NOT create a group member', () => {
      return request(server)
        .post('/groups/' + otherGroup.id + '/members')
        .send(user)
        .set('From', user.email)
        .set('Accept', 'application/json')
        .expect(404)
    })

    it('should create a group member', () => {
      var params = _.pick(user, 'email', 'fullName')
      return request(server)
        .post('/groups/' + group.id + '/members')
        .send(params)
        .set('From', member.email)
        .set('Accept', 'application/json')
        .expect(200)
    })
  })

  describe('DELETE /groups/:groupId/members/:personId', () => {
    it('should return 401', () => {
      return request(server)
        .del('/groups/' + group.id + '/members/' + member.id)
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should remove a group member', () => {
      return request(server)
        .del('/groups/' + group.id + '/members/' + friend.id)
        .set('From', member.email)
        .set('Accept', 'application/json')
        .expect(200)
    })
  })
})
