/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const request = require('supertest')
const {expect} = require('chai')
const server = require('../server')
const Factory = require('../factory')

describe('Organizations Controller', () => {
  var admin
  var member
  var organization

  beforeEach(async () => {
    organization = await Factory.create('organization')
    member = await Factory.createPerson(organization, {member: true})
    await Factory.createPerson(organization, {member: false})
    await Factory.createPerson(organization, {member: true, archived: true})

    admin = await Factory.createUser({admin: true})
    await Factory.createUser()
  })

  describe('GET /organizations @cleandb', () => {
    it('should return 401', () => {
      return request(server)
        .get('/organizations')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return organizations', () => {
      return request(server)
        .get('/organizations')
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).expect(function (res) {
          var organizations = res.body
          expect(organizations).to.have.length(1)
        })
    })
  })

  describe('POST /organizations', () => {
    it('should return 401', () => {
      return request(server)
        .post('/organizations')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should create an organization', () => {
      return request(server)
        .post('/organizations')
        .send({
          name: 'New Organization',
          owner: {
            email: 'owner@organization.com',
            fullName: 'Owner Person'
          }
        })
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var json = res.body
          expect(json.name).to.equal('New Organization')
          expect(json.groups).to.have.length(1)
        })
    })
  })

  describe('GET /organizations/:id', () => {
    it('should return 401', () => {
      return request(server)
        .get('/organizations/' + organization.id)
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return an organization', () => {
      return request(server)
        .get('/organizations/' + organization.id)
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var json = res.body
          expect(json.id).to.equal(organization.id)
          expect(json.peopleCount).to.equal(2)
          expect(json.membersCount).to.equal(1)
          expect(json.friendsCount).to.equal(1)
        })
    })
  })

  describe('PATCH /organizations/:id', () => {
    it('should return 401', () => {
      return request(server)
        .patch('/organizations/' + organization.id)
        .send({name: 'New Name'})
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return 403', () => {
      return request(server)
        .patch('/organizations/' + organization.id)
        .send({name: 'New Name'})
        .set('From', member.email)
        .set('Accept', 'application/json')
        .expect(403)
    })

    it('should update a user', () => {
      return request(server)
        .patch('/organizations/' + organization.id)
        .send({name: 'New Name'})
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var json = res.body
          expect(json.name).to.equal('New Name')
        })
    })
  })

  describe('DELETE /organizations/:id', () => {
    it('should return 401', () => {
      return request(server)
        .del('/organizations/' + organization.id)
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should delete an organization', () => {
      return request(server)
        .del('/organizations/' + organization.id)
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200)
    })
  })
})
