/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const path = require('path')
const request = require('supertest')
const {expect} = require('chai')
const server = require('../server')
const Factory = require('../factory')
const {Email, Person} = require('../rbac')

describe('People Controller', () => {
  var admin
  var user
  var organization
  var otherOrganization
  var group
  var person
  var person2
  var person3

  beforeEach(async () => {
    organization = await Factory.create('organization')
    otherOrganization = await Factory.create('organization')

    person = await Factory.createPerson(organization, {email: 'sam@test.com', member: true})
    person2 = await Factory.createPerson(organization, {member: true})
    person3 = await Factory.createPerson(organization)
    await Factory.createPerson(organization, {member: true, archived: true})

    group = await Factory.createGroup(organization, {name: 'Admin'})
    await Factory.addGroupMember(group, person)
    await Factory.addGroupMember(group, person3)

    admin = await Factory.createUser({admin: true})
    user = await Factory.createUser()
  })

  describe('GET /people @cleandb', () => {
    it('should return people (group)', () => {
      return request(server)
        .get('/people?friends=1&group=' + group.id)
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var people = res.body
          expect(people).to.have.length(2)
        })
    })

    it('should return people (organization)', () => {
      return request(server)
        .get('/people?friends=1&organization=' + organization.id)
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var people = res.body
          expect(people).to.have.length(3)
        })
    })

    it('should return people (admin)', () => {
      return request(server)
        .get('/people?friends=1')
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var people = res.body
          expect(people).to.have.length(3)
        })
    })
  })

  describe('POST /organizations/:id/people', () => {
    it('should return 401', () => {
      return request(server)
        .post('/organizations/' + organization.id + '/people')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should NOT create an organization member', () => {
      return request(server)
        .post('/organizations/' + otherOrganization.id + '/people')
        .send(user)
        .set('From', user.email)
        .set('Accept', 'application/json')
        .expect(404)
    })

    it('should create an organization person', () => {
      var params = _.pick(user, 'email', 'fullName')
      return request(server)
        .post('/organizations/' + organization.id + '/people')
        .send(params)
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200)
    })
  })

  describe('POST /organizations/:id/people/upload', () => {
    const csv = path.resolve(__dirname, '..', 'data', 'members.csv')

    it('should return 404', () => {
      return request(server)
        .post('/organizations/' + organization.id + '/people/upload')
        .set('From', user.email)
        .set('Accept', 'application/json')
        .expect(404)
    })

    it('should import people', () => {
      return request(server)
        .post('/organizations/' + organization.id + '/people/upload')
        .attach('file', csv)
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          expect(res.body).to.have.length(2)
        })
    })
  })

  describe('POST /organizations/:id/people/download', () => {
    it('should return 404', () => {
      return request(server)
        .get('/organizations/' + organization.id + '/people/download')
        .set('From', user.email)
        .expect(404)
    })

    it('should export people', () => {
      return request(server)
        .get('/organizations/' + organization.id + '/people/download')
        .set('From', person.email)
        .expect(200).then(function (res) {
          // Parse all lines
          var lines = _.map(res.text.trim().split('\n'), function (line) {
            return line.replace('"', '')
          })
          expect(lines).to.have.length(3)

          // Parse headings
          var headings = lines[0].split(',')
          expect(headings).to.have.length(3)
        })
    })
  })

  describe('PATCH /people/:id', () => {
    it('should return 401', () => {
      return request(server)
        .patch('/people/' + person2.id)
        .send({fullName: 'New Name'})
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should update their name', () => {
      return request(server)
        .patch('/people/' + person2.id)
        .send({fullName: 'New Name'})
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          return Person.findById(person2.id).then(function (person) {
            expect(person.fullName).to.eql('New Name')
          })
        })
    })

    it('should update their email', () => {
      return request(server)
        .patch('/people/' + person2.id)
        .send({email: 'new@new.com'})
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          return Person.findById(person2.id).then(function (person) {
            return Email.loadEmail(person)
          }).then(function (person) {
            expect(person.email).to.eql('new@new.com')
          })
        })
    })
  })

  describe('DELETE /people/:id', () => {
    it('should return 401', () => {
      return request(server)
        .del('/people/' + person2.id)
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should remove an organization member', () => {
      return request(server)
        .del('/people/' + person2.id)
        .set('From', person.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          return Person.findById(person2.id).then(function (person) {
            expect(person).to.be.undefined
          })
        })
    })
  })
})
