/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const request = require('supertest')
const {expect} = require('chai')
const server = require('../server')
const Factory = require('../factory')
const {User} = require('../rbac')

describe('Users Controller', () => {
  var admin
  var nonAdmin
  var organization
  var person

  beforeEach(async () => {
    admin = await Factory.createUser({email: 'sam@sam.com', admin: true})
    nonAdmin = await Factory.createUser({email: 'tom@tom.com'})

    organization = await Factory.create('organization')
    person = await Factory.createPerson(organization, {member: true})
  })

  describe('POST /users/validate', () => {
    it('should fail an existing email', () => {
      return request(server)
        .post('/users/validate')
        .query({email: nonAdmin.email})
        .expect(200).then(function (res) {
          expect(res.body.email).to.be.false
        })
    })

    it('should pass for an available email', () => {
      return request(server)
        .post('/users/validate')
        .query({email: 'email@available.com'})
        .expect(200).then(function (res) {
          expect(res.body.email).to.be.true
        })
    })
  })

  describe('GET /users @cleandb', () => {
    it('should return 401', () => {
      return request(server)
        .get('/users')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return 403', () => {
      return request(server)
        .get('/users')
        .set('From', nonAdmin.email)
        .set('Accept', 'application/json')
        .expect(403)
    })

    it('should return all users', () => {
      return request(server)
        .get('/users')
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var users = res.body
          expect(users).to.have.length(3)

          var emails = _.pluck(users, 'email')
          expect(emails).to.eql([admin.email, nonAdmin.email, person.email])

          var user = _.findWhere(users, {id: person.user.id})
          var peopleIds = _.pluck(user.people, 'id')
          expect(peopleIds).to.eql([person.id])
        })
    })

    it('should return a subset of users', () => {
      return request(server)
        .get('/users?search=sa')
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var users = res.body
          expect(users).to.have.length(1)
        })
    })
  })

  describe('GET /users/:id', () => {
    it('should return 401', () => {
      return request(server)
        .get('/users/' + person.user.id)
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return 403', () => {
      return request(server)
        .get('/users/' + person.user.id)
        .set('From', nonAdmin.email)
        .set('Accept', 'application/json')
        .expect(403)
    })

    it('should return a user', () => {
      return request(server)
        .get('/users/' + person.user.id)
        .set('From', admin.email)
        .set('Accept', 'application/json')
        .expect(200).then(function (res) {
          var json = res.body
          expect(json.email).to.eql(person.email)

          var peopleIds = _.pluck(json.people, 'id')
          expect(peopleIds).to.eql([person.id])
        })
    })
  })

  describe('POST /users/reset @cleandb', () => {
    it('should return 401', () => {
      return request(server)
        .post('/users/reset')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return 403', () => {
      return request(server)
        .post('/users/reset')
        .set('From', nonAdmin.email)
        .set('Accept', 'application/json')
        .expect(403)
    })

    it('should reset all users', () => {
      return request(server)
        .post('/users/reset')
        .set('From', admin.email)
        .expect(200).then(() => {
          return User.query().then(function (users) {
            _.each(users, function (user) {
              expect(user.resetSessionAt).to.not.be.null
            })
          })
        })
    })

    it('should reset some users', () => {
      return request(server)
        .post('/users/reset')
        .send({search: 'sam'})
        .set('From', admin.email)
        .expect(200).then(() => {
          return User.query().whereNotNull('reset_session_at').then(function (users) {
            var userIds = _.pluck(users, 'id')
            expect(userIds).to.eql([admin.id])
          })
        })
    })
  })

  describe('POST /users/:userId/reset @cleandb', () => {
    it('should return 401', () => {
      return request(server)
        .post('/users/' + person.user.id + '/reset')
        .set('Accept', 'application/json')
        .expect(401)
    })

    it('should return 403', () => {
      return request(server)
        .post('/users/' + person.user.id + '/reset')
        .set('From', nonAdmin.email)
        .set('Accept', 'application/json')
        .expect(403)
    })

    it('should reset the user', () => {
      return request(server)
        .post('/users/' + person.user.id + '/reset')
        .set('From', admin.email)
        .expect(200).then(() => {
          return User.query().whereNotNull('reset_session_at').then(function (users) {
            var userIds = _.pluck(users, 'id')
            expect(userIds).to.eql([person.user.id])
          })
        })
    })
  })
})
