/* global describe it */
const _ = require('underscore')
const moment = require('moment')
const fs = require('fs')
const path = require('path')
const {expect} = require('chai')
const Rbac = require('./rbac')
const csv = require('../server/csv')

function createStream (filename) {
  const csv = path.resolve(__dirname, 'data', filename)
  return fs.createReadStream(csv)
}

describe('csv', () => {
  describe('loadStream', () => {
    it('should load headers and rows', () => {
      var stream = createStream('members.csv')
      return csv.loadStream(stream).then(function (data) {
        expect(data.headers).to.eql(['Email', 'Name', 'Group'])
        expect(data.rows).to.eql([
          ['Sam@test.com', 'Sam Duvall', 'Group #1'],
          ['maria@test.com', 'Maria Duvall']
        ])
      })
    })
  })

  describe('parseData', () => {
    var email = {
      name: 'Email',
      field: 'email',
      regEx: /email?/i,
      type: 'email',
      required: true,
      unique: true
    }

    var name = {
      name: 'Name',
      field: 'name',
      regEx: /name?/i,
      required: true
    }

    var hireDate = {
      name: 'Hire Date',
      field: 'hireDate',
      regEx: /hired? date?/i,
      type: 'date'
    }

    var groups = {
      name: 'Groups',
      field: 'groups',
      regEx: /groups?/i,
      type: 'array',
      exclude: ['Admin']
    }

    var tests = [{
      name: 'should lower case emails',
      headers: ['Email'],
      rows: [
        ['A@a.cOm']
      ],
      schema: [email],
      entries: [
        {email: 'a@a.com'}
      ]
    }, {
      name: 'should validate required columns (headers)',
      headers: [],
      rows: [
        []
      ],
      schema: [email],
      error: 'Missing required column: Email'
    }, {
      name: 'should validate required columns (rows)',
      headers: ['Email'],
      rows: [
        []
      ],
      schema: [email],
      error: 'Missing Email on line 2'
    }, {
      name: 'should validate unique columns',
      headers: ['Email'],
      rows: [
        ['a@a.com'],
        ['b@b.com'],
        ['A@a.com']
      ],
      schema: [email],
      error: 'Duplicate Email: a@a.com'
    }, {
      name: 'should allow unique vs required columns',
      headers: ['Email', 'Name'],
      rows: [
        ['sam@rocketwhale.com', 'Sammy D'],
        ['sam@sam.com', 'Sammy D']
      ],
      schema: [email, name],
      entries: [
        {email: 'sam@rocketwhale.com', name: 'Sammy D'},
        {email: 'sam@sam.com', name: 'Sammy D'}
      ]
    }, {
      name: 'should disallow invalid date',
      headers: ['Hire Date'],
      rows: [
        ['2x16-10-28']
      ],
      schema: [hireDate],
      error: 'Hire Date must have valid format (YYYY-MM-DD or M/D/YYYY). Found invalid value: 2x16-10-28.'
    }, {
      name: 'should allow valid date (YYYY-MM-DD)',
      headers: ['Hire Date'],
      rows: [
        ['2016-10-28']
      ],
      schema: [hireDate],
      entries: [
        {hireDate: moment('2016-10-28').toDate()}
      ]
    }, {
      name: 'should allow valid date (MM/DD/YYYY)',
      headers: ['Hire Date'],
      rows: [
        ['08/04/1981']
      ],
      schema: [hireDate],
      entries: [
        {hireDate: moment('1981-08-04').toDate()}
      ]
    }, {
      name: 'should allow valid date (M/D/YYYY)',
      headers: ['Hire Date'],
      rows: [
        ['8/4/1981']
      ],
      schema: [hireDate],
      entries: [
        {hireDate: moment('1981-08-04').toDate()}
      ]
    }, {
      name: 'should default an array column',
      headers: ['Groups'],
      rows: [
        []
      ],
      schema: [groups],
      entries: [
        {groups: []}
      ]
    }, {
      name: 'should allow multiple entries on an array column (duplicate header)',
      headers: ['Groups', 'Group'],
      schema: [groups],
      rows: [
        ['Alpha', 'Beta']
      ],
      entries: [
        {groups: ['Alpha', 'Beta']}
      ]
    }, {
      name: 'should allow multiple entries on an array column (empty header)',
      headers: ['Unknown', 'Groups', ''],
      schema: [groups],
      rows: [
        ['Wrong', 'Alpha', 'Beta']
      ],
      entries: [
        {groups: ['Alpha', 'Beta']}
      ]
    }, {
      name: 'should allow multiple entries on an array column (no header)',
      headers: ['Unknown', 'Groups'],
      schema: [groups],
      rows: [
        ['Wrong', 'Alpha', 'Beta']
      ],
      entries: [
        {groups: ['Alpha', 'Beta']}
      ]
    }, {
      name: 'should allow multiple entries on an array column (truncated)',
      headers: ['Unknown', 'Groups', '', 'Other'],
      schema: [groups],
      rows: [
        ['Wrong', 'Alpha', 'Beta', 'Wrong']
      ],
      entries: [
        {groups: ['Alpha', 'Beta']}
      ]
    }, {
      name: 'should exclude entries on an array column tacos',
      headers: ['Unknown', 'Groups', ''],
      schema: [groups],
      rows: [
        ['Wrong', 'Alpha', 'AdMin']
      ],
      entries: [
        {groups: ['Alpha']}
      ]
    }]

    _.each(tests, function (test) {
      it('should ' + test.name, () => {
        var data = _.pick(test, 'headers', 'rows')
        if (test.error) {
          var parseData = _.partial(csv.parseData, data, test.schema)
          expect(parseData).to.throw(csv.Error, test.error)
        } else {
          var entries = csv.parseData(data, test.schema)
          expect(entries).to.eql(test.entries)
        }
      })
    })
  })

  describe('parseStream', () => {
    it('should map the rows to fields', () => {
      var stream = createStream('members.csv')
      return csv.parseStream(stream, Rbac.Person.csvSchema).then(function (entries) {
        expect(entries).to.eql([
          {email: 'sam@test.com', fullName: 'Sam Duvall', groups: ['Group #1']},
          {email: 'maria@test.com', fullName: 'Maria Duvall', groups: []}
        ])
      })
    })
  })
})
