const isArray = require('lodash/isArray')
const omit = require('lodash/omit')
const {Factory} = require('rosie')
const Rbac = require('./rbac')
require('./database')

Factory.define = function (name, schema) {
  var factory = new Factory()
  factory.schema = schema
  this.factories[name] = factory
  return factory
}

Factory.create = async function (name, params) {
  const build = (attr) => this.build(name, attr)
  const attr = isArray(params) ? params.map(build) : build(params)
  const factory = this.factories[name]
  return factory.schema.create(attr)
}

Factory.define('email', Rbac.Email)
  .sequence('email', function (i) { return 'email-' + i + '@email.com' })

Factory.define('person', Rbac.Person)
  .sequence('fullName', function (i) { return 'Person #' + i })

Factory.define('group', Rbac.Group)
  .sequence('name', function (i) { return 'Group #' + i })

Factory.define('group.member', Rbac.Group.Member)

Factory.define('privilege', Rbac.Privilege)
  .sequence('name', function (i) { return 'Privilege #' + i })

Factory.define('organization', Rbac.Organization)
  .sequence('name', function (i) { return 'Organization #' + i })
  .sequence('domain', function (i) { return 'domain-' + i + '.com' })

Factory.define('permission', Rbac.Permission)

Factory.define('user', Rbac.User)
  .sequence('fullName', function (i) { return 'User #' + i })

// Create a person (ensures user exists)
Factory.createPerson = async function (organization, params) {
  params = params || {}

  // Create associations
  var email = await Factory.createEmail(params.email)
  var user = await Factory.createUser({email: email.email})

  // Format params
  params = omit(params, 'email')
  params.emailId = email.id
  params.organizationId = organization.id

  // Create person
  var person = await Factory.create('person', params)
  person.email = email.email
  person.user = user
  return person
}

// Create email (optional email field)
Factory.createEmail = async function (email) {
  if (!email) return Factory.create('email')
  const row = await Rbac.Email.findByEmail(email)
  return row || Factory.create('email', {email: email})
}

// Create a group
Factory.createGroup = async function (organization, params) {
  params = params || {}
  params.organizationId = organization.id
  return Factory.create('group', params)
}

// Add a group member
Factory.addGroupMember = async function (group, person) {
  return Factory.create('group.member', {
    groupId: group.id,
    personId: person.id
  })
}

// Create an email user
Factory.createUser = async function (params) {
  params = params || {}

  // Create associations
  var email = await Factory.createEmail(params.email)

  // Format params
  params = omit(params, 'email')
  params.emailId = email.id

  // Create user
  var user = await Factory.create('user', params)
  user.email = email.email
  return user
}

Factory.createPermission = async function (privilege, organization, expression, resourceType, resourceId) {
  return Factory.create('permission', {
    privilegeId: privilege.id,
    organizationId: organization.id,
    expression: expression,
    resourceType: resourceType,
    resourceId: resourceId
  })
}

module.exports = Factory
