/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const {expect} = require('chai')
const Factory = require('../factory')
const {Email} = require('../rbac')

describe('Email', () => {
  var email

  beforeEach(async () => {
    email = await Factory.createEmail('sam@sam.com')
  })

  describe('whereSearch', () => {
    it('should only find users with search text', () => {
      return Email.query().whereSearch('sa').then(function (emails) {
        emails = _.pluck(emails, 'email')
        expect(emails).to.eql(['sam@sam.com'])
      })
    })
  })

  describe('findId', () => {
    it('should find a user by email', () => {
      return Email.findId(email.email).then(function (id) {
        expect(id).to.eql(email.id)
      })
    })
  })

  describe('findOrCreateId', () => {
    it('should find an existing email', () => {
      return Email.findOrCreateId(email.email).then(function (id) {
        expect(id).to.eql(email.id)
      })
    })

    it('should create a email', () => {
      return Email.findOrCreateId('test@test.com').then(function (id) {
        expect(id).to.not.eql(email.id)
      })
    })
  })

  describe('bulkFindOrCreateByEmail', () => {
    beforeEach(async () => {
      await Factory.createEmail('test@test.com')
    })

    it('should create multiple emails in order', () => {
      var before = [undefined, 'Tom@Tom.com', undefined, 'test@test.com', 'Sam@Sam.com']
      var after = ['sam@sam.com', 'test@test.com', 'tom@tom.com']
      return Email.bulkFindOrCreateByEmail(before).then(function (rows) {
        var emails = _.pluck(rows, 'email')
        expect(emails).to.eql(after)
      })
    })
  })

  describe('loadEmail', () => {
    var organization
    var person
    var person2

    beforeEach(async () => {
      organization = await Factory.create('organization')
      person = await Factory.createPerson(organization, {member: true})
      person2 = await Factory.createPerson(organization, {member: true})
    })

    it('should map emailId to emails', () => {
      var objects = [
        {emailId: 88888},
        {emailId: person2.emailId},
        {emailId: undefined},
        {emailId: person.emailId},
        {emailId: 99999}
      ]

      return Email.loadEmail(objects).then(() => {
        var expected = [undefined, person2.email, undefined, person.email, undefined]
        var emails = _.pluck(objects, 'email')
        expect(emails).to.eql(expected)
      })
    })
  })

  describe('loadEmailId', () => {
    var organization
    var person
    var person2

    beforeEach(async () => {
      organization = await Factory.create('organization')
      person = await Factory.createPerson(organization, {member: true})
      person2 = await Factory.createPerson(organization, {member: true})
    })

    it('should map emailId to emails', () => {
      var objects = [
        {email: 'New@New.com'},
        {email: person2.email},
        {email: undefined},
        {email: person.email},
        {email: 'Last@Last.com'}
      ]

      return Email.loadEmailId(objects).then(() => {
        expect(objects[0].emailId).to.not.be.undefined
        expect(objects[1].emailId).to.eql(person2.emailId)
        expect(objects[2].emailId).to.be.undefined
        expect(objects[3].emailId).to.eql(person.emailId)
        expect(objects[4].emailId).to.not.be.undefined
      })
    })
  })
})
