/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const {expect} = require('chai')
const Factory = require('../factory')
const {Group, Organization, Permission} = require('../rbac')

describe('Group', () => {
  var organization
  var adminGroup
  var otherGroup
  var archivedGroup
  var person
  var person3

  beforeEach(async () => {
    organization = await Factory.create('organization')
    person = await Factory.createPerson(organization, {member: true})
    await Factory.createPerson(organization, {member: true})
    person3 = await Factory.createPerson(organization)

    adminGroup = await Factory.create('group', {organizationId: organization.id, name: 'Admin'})
    otherGroup = await Factory.create('group', {organizationId: organization.id, name: 'Other'})
    archivedGroup = await Factory.create('group', {organizationId: organization.id, name: 'Archived', archived: true})
    await Factory.addGroupMember(adminGroup, person)
    await Factory.addGroupMember(adminGroup, person3)
  })

  describe('whereSearch', () => {
    it('should only find groups with search text', () => {
      return Group.query().whereSearch('oth').then(function (groups) {
        var names = _.pluck(groups, 'name')
        expect(names).to.eql(['Other'])
      })
    })
  })

  describe('whereEmpty @cleandb', () => {
    it('should only return groups with no members', () => {
      return Group.query().whereEmpty().then(function (groups) {
        var groupIds = _.pluck(groups, 'id')
        expect(groupIds).to.eql([otherGroup.id, archivedGroup.id])
      })
    })
  })

  describe('wherePersonId', () => {
    it('should only return groups the person is on', () => {
      return Group.query().wherePersonId(person.id).then(function (groups) {
        var groupIds = _.pluck(groups, 'id')
        expect(groupIds).to.eql([adminGroup.id])
      })
    })
  })

  describe('findByName', () => {
    it('should NOT find an non-existing group', () => {
      return Group.findByName(organization.id, 'unknown').then(function (group) {
        expect(group).to.be.undefined
      })
    })

    it('should find an existing group case-insensitive', () => {
      return Group.findOrCreateByName(organization.id, 'admin').then(function (group) {
        expect(group.id).to.eql(adminGroup.id)
        expect(group.name).to.eql(adminGroup.name)
      })
    })
  })

  describe('findOrCreateByName', () => {
    it('should find an existing group case-insensitive', () => {
      return Group.findOrCreateByName(organization.id, 'admin').then(function (group) {
        expect(group.id).to.eql(adminGroup.id)
        expect(group.name).to.eql(adminGroup.name)
        expect(group.archived).to.be.false
      })
    })

    it('should create a new group', () => {
      return Group.findOrCreateByName(organization.id, 'New Group').then(function (group) {
        expect(group.id).to.not.be.undefined
        expect(group.name).to.eql('New Group')
        expect(group.archived).to.be.false
      })
    })

    it('should unarchive an archived group', () => {
      return Group.findOrCreateByName(organization.id, 'Archived').then(function (group) {
        expect(group.id).to.eql(archivedGroup.id)
        expect(group.name).to.eql(archivedGroup.name)
        expect(group.archived).to.be.false
      })
    })
  })

  describe('isNameEqual', () => {
    it('should be equal case-insensitive', () => {
      var equal = Group.isNameEqual(adminGroup, 'admin')
      expect(equal).to.be.true
    })

    it('should NOT be equal if different', () => {
      var equal = Group.isNameEqual(adminGroup, 'other')
      expect(equal).to.be.false
    })
  })

  describe('whereQuery @cleandb', () => {
    it('should return active groups on an organization', () => {
      return Group.query().whereQuery({organizationId: organization.id}).then(function (groups) {
        var ids = _.pluck(groups, 'id')
        expect(ids).to.eql([adminGroup.id, otherGroup.id])
      })
    })

    it('should exclude certain groups', () => {
      return Group.query().whereQuery({organizationId: organization.id, exclude: adminGroup.id}).then(function (groups) {
        var ids = _.pluck(groups, 'id')
        expect(ids).to.eql([otherGroup.id])
      })
    })

    it('should return all groups', () => {
      return Group.query().whereQuery({organizationId: organization.id, archived: 'true'}).then(function (groups) {
        var ids = _.pluck(groups, 'id')
        expect(ids).to.eql([adminGroup.id, otherGroup.id, archivedGroup.id])
      })
    })
  })

  describe('bulkCreate', () => {
    it('should only find groups', () => {
      return Group.bulkCreate(organization, [adminGroup.name, otherGroup.name]).then(function (groups) {
        var groupIds = _.pluck(groups, 'id')
        expect(groupIds).to.eql([adminGroup.id, otherGroup.id])
      })
    })

    it('should only create groups', () => {
      var names = ['Another Group', 'New Group']
      return Group.bulkCreate(organization, names).then(function (groups) {
        var groupNames = _.pluck(groups, 'name')
        expect(groupNames).to.eql(names)
      })
    })

    it('should only unarchive groups', () => {
      return Group.bulkCreate(organization, [archivedGroup.name]).then(function (groups) {
        var groupIds = _.pluck(groups, 'id')
        expect(groupIds).to.eql([archivedGroup.id])

        var archived = _.pluck(groups, 'archived')
        expect(archived).to.eql([false])
      })
    })

    it('should handle a mixed bag', () => {
      return Group.bulkCreate(organization, ['admin', 'archived', 'DC']).then(function (groups) {
        // Retain existing group
        let group = _.findWhere(groups, {id: adminGroup.id})
        expect(group).not.to.be.undefined
        expect(group.name).to.eql(adminGroup.name)
        expect(group.archived).to.be.false

        // Unarchive archived group
        group = _.findWhere(groups, {id: archivedGroup.id})
        expect(group).not.to.be.undefined
        expect(group.name).to.eql(archivedGroup.name)
        expect(group.archived).to.be.false

        // Create new group
        group = _.findWhere(groups, {name: 'DC'})
        expect(group).not.to.be.undefined
        expect(group.name).to.eql('DC')
        expect(group.archived).to.be.false
      })
    })
  })

  describe('archive', () => {
    var expression
    var privilege
    var permission

    beforeEach(async () => {
      expression = 'O' + organization.id + '||' + 'G' + adminGroup.id
      privilege = await Factory.create('privilege')
      permission = await Factory.createPermission(privilege, organization, expression)
    })

    it('should NOT remove the group from the organization', () => {
      return Group.archive(adminGroup).then(() => {
        return Organization.load(organization, 'groups')
      }).then(function (organization) {
        // Existing
        let groups = _.where(organization.groups, {archived: false})
        let names = _.pluck(groups, 'name')
        expect(names).to.eql([otherGroup.name])
        // Archived
        groups = _.where(organization.groups, {archived: true})
        names = _.pluck(groups, 'name')
        expect(names).to.eql([adminGroup.name, archivedGroup.name])
      })
    })

    it('should remove all group members', () => {
      return Group.archive(adminGroup).then(() => {
        return Group.Member.query().where('groupId', adminGroup.id)
      }).then(function (members) {
        expect(members).to.be.empty
      })
    })

    it('should remove from permissions', () => {
      return Group.archive(adminGroup).then(() => {
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql('O' + organization.id)
      })
    })
  })

  describe('purge', () => {
    var expression
    var privilege
    var permission

    beforeEach(async () => {
      expression = 'O' + organization.id + '||' + 'G' + adminGroup.id
      privilege = await Factory.create('privilege')
      permission = await Factory.createPermission(privilege, organization, expression)
    })

    it('should remove the group from the organization', () => {
      return Group.purge(adminGroup).then(() => {
        return Organization.load(organization, 'groups')
      }).then(function (organization) {
        // Existing
        let groups = _.where(organization.groups, {archived: false})
        let names = _.pluck(groups, 'name')
        expect(names).to.eql([otherGroup.name])
        // Archived
        groups = _.where(organization.groups, {archived: true})
        names = _.pluck(groups, 'name')
        expect(names).to.eql([archivedGroup.name])
      })
    })

    it('should remove all group members', () => {
      return Group.purge(adminGroup).then(() => {
        return Group.Member.query().where('groupId', adminGroup.id)
      }).then(function (members) {
        expect(members).to.be.empty
      })
    })

    it('should remove from permissions', () => {
      return Group.purge(adminGroup).then(() => {
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql('O' + organization.id)
      })
    })
  })

  describe('Member', () => {
    describe('removeFromGroup', () => {
      it('remove a person from a group', () => {
        return Group.Member.removeFromGroup(adminGroup, person3).then(() => {
          return Group.Member.whereGroupId(adminGroup.id)
        }).then(function (members) {
          var personIds = _.pluck(members, 'personId')
          expect(personIds).to.eql([person.id])
        })
      })
    })

    describe('bulkCreate', () => {
      var member = {email: 'member@member.com', fullName: 'Member #1'}
      var member2 = {email: 'member2@member.com', fullName: 'Member #2'}

      it('should NOT create a new member', () => {
        var params = _.pick(person, 'email', 'fullName')
        return Group.Member.bulkCreate(adminGroup, params).then(function (people) {
          var emails = _.pluck(people, 'email')
          expect(emails).to.eql([person.email])
          return Group.Member.whereGroupId(adminGroup.id)
        }).then(function (members) {
          var personIds = _.pluck(members, 'personId')
          expect(personIds).to.eql([person.id, person3.id])
        })
      })

      it('should create a new member', () => {
        return Group.Member.bulkCreate(adminGroup, member).then(function (people) {
          var newPerson = _.findWhere(people, {email: member.email})
          var emails = _.pluck(people, 'email')
          expect(emails).to.eql([member.email])
          return Group.Member.whereGroupId(adminGroup.id).then(function (members) {
            var personIds = _.pluck(members, 'personId')
            expect(personIds).to.eql([person.id, person3.id, newPerson.id])
          })
        })
      })

      it('should create new members', () => {
        return Group.Member.bulkCreate(adminGroup, [member, member2]).then(function (people) {
          var newPerson = _.findWhere(people, {email: member.email})
          var newPerson2 = _.findWhere(people, {email: member2.email})
          var emails = _.pluck(people, 'email')
          expect(emails).to.eql([member.email, member2.email])
          return Group.Member.whereGroupId(adminGroup.id).then(function (members) {
            var personIds = _.pluck(members, 'personId')
            expect(personIds).to.eql([person.id, person3.id, newPerson.id, newPerson2.id])
          })
        })
      })
    })
  })
})
