/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const Promise = require('bluebird')
const {expect} = require('chai')
const Factory = require('../factory')
const {Group, Organization, Permission, Person} = require('../rbac')

describe('Organization', () => {
  var organization
  var organization2
  var group
  var admin
  var person
  var person3
  var person4

  beforeEach(async () => {
    admin = await Factory.createUser({admin: true})
    await Factory.createUser({email: 'sam@test.com'})

    organization = await Factory.create('organization')
    organization2 = await Factory.create('organization', {name: 'Other'})

    person = await Factory.createPerson(organization, {member: true})
    await Factory.createPerson(organization, {member: true})
    person3 = await Factory.createPerson(organization)
    person4 = await Factory.createPerson(organization, {member: true, archived: true})

    group = await Factory.createGroup(organization, {name: 'Admin'})
    await Factory.addGroupMember(group, person)
    await Factory.addGroupMember(group, person3)
  })

  describe('createBy @cleandb', () => {
    beforeEach(async () => {
      await Factory.create('privilege')
      await Factory.create('privilege')
    })

    it('should create a new organization', () => {
      return Organization.createBy({
        name: 'New Organization',
        owner: {
          email: 'Owner@organization.com',
          fullName: 'Owner McOwnerson'
        }
      }).then(function (organization) {
        return Promise.all([
          Person.where({organizationId: organization.id}).joinEmail(),
          Organization.loadGroups(organization),
          Permission.query().whereOrganizationId(organization.id)
        ])
      }).spread(function (people, organization, permissions) {
        expect(people).to.have.length(1)
        expect(permissions).to.have.length(2)

        var owner = _.findWhere(people, {email: 'owner@organization.com'})
        expect(owner.fullName).to.eql('Owner McOwnerson')
        expect(owner.member).to.be.true

        var group = _.findWhere(organization.groups, {name: 'Admin'})
        expect(group.peopleCount).to.eql(1)
      })
    })
  })

  describe('whereUserCan @cleandb', () => {
    it('should not find organizations if no user', () => {
      return Organization.whereUserCan().then(function (organizations) {
        expect(organizations).to.be.empty
      })
    })

    it('should find all organizations the user is an admin', () => {
      return Organization.whereUserCan(admin).then(function (organizations) {
        var organizationIds = _.pluck(organizations, 'id')
        expect(organizationIds).to.eql([organization.id, organization2.id])
      })
    })

    it('should find organizations the user is a member of', () => {
      return Organization.whereUserCan(person.user).then(function (organizations) {
        var organizationIds = _.pluck(organizations, 'id')
        expect(organizationIds).to.eql([organization.id])
      })
    })

    it('should find organizations the user is a friend of', () => {
      return Organization.whereUserCan(person3.user).then(function (organizations) {
        var organizationIds = _.pluck(organizations, 'id')
        expect(organizationIds).to.eql([organization.id])
      })
    })

    it('should NOT find organizations where the user has been archived', () => {
      return Organization.whereUserCan(person4.user).then(function (organizations) {
        var organizationIds = _.pluck(organizations, 'id')
        expect(organizationIds).to.be.empty
      })
    })
  })

  describe('whereSearch @cleandb', () => {
    it('should only find organizations with search text', () => {
      return Organization.query().whereSearch('oth').then(function (organizations) {
        var names = _.pluck(organizations, 'name')
        expect(names).to.eql(['Other'])
      })
    })
  })

  describe('loadGroups', () => {
    it('should load groups and their count', () => {
      return Organization.loadGroups(organization).then(function (organization) {
        group = _.findWhere(organization.groups, {id: group.id})
        expect(group).to.not.be.undefined
        expect(group.peopleCount).to.eql(2)
      })
    })
  })

  describe('removeById', () => {
    it('should remove all groups/people/permissions', () => {
      return Organization.removeById(organization.id).then(() => {
        return Promise.all([
          Organization.where({id: organization.id}),
          Group.where({organization_id: organization.id}),
          Group.Member.where({group_id: group.id}),
          Permission.where({organization_id: organization.id}),
          Person.where({organization_id: organization.id})
        ])
      }).spread(function (organizations, groups, groupMembers, permissions, people) {
        expect(organizations).to.be.empty
        expect(groups).to.be.empty
        expect(groupMembers).to.be.empty
        expect(permissions).to.be.empty
        expect(people).to.be.empty
      })
    })
  })
})
