/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const Promise = require('bluebird')
const {expect} = require('chai')
const Factory = require('../factory')
const {Organization, Permission} = require('../rbac')

describe('Permission', () => {
  var organization
  var organization2
  var person
  var person2
  var person3
  var person4
  var person5

  var editOrg
  var viewPerms
  var editPerms

  beforeEach(async () => {
    organization = await Factory.create('organization')
    organization2 = await Factory.create('organization')

    person = await Factory.createPerson(organization, {member: true})
    person2 = await Factory.createPerson(organization, {member: true})
    person3 = await Factory.createPerson(organization, {member: false})
    person4 = await Factory.createPerson(organization, {member: true, archived: true})
    person5 = await Factory.createPerson(organization2, {member: true})

    editOrg = await Factory.create('privilege', {name: 'Edit Organization'})
    editOrg.permission = await Factory.createPermission(editOrg, organization, 'P' + person.id)

    viewPerms = await Factory.create('privilege', {name: 'View Permissions'})
    viewPerms.permission = await Factory.createPermission(viewPerms, organization, 'O' + organization.id)
    viewPerms.permission2 = await Factory.createPermission(viewPerms, organization2, 'O' + organization2.id)

    editPerms = await Factory.create('privilege', {name: 'Edit Permissions'})
    editPerms.permission = await Factory.createPermission(editPerms, organization, 'P' + person2.id, 'permissions', editOrg.id)
  })

  describe('whereUser', () => {
    it('scope by privilege', () => {
      var tests = [
        {user: person.user, permissions: [viewPerms.permission]},
        {user: person2.user, permissions: [viewPerms.permission]},
        {user: person3.user, permissions: []},
        {user: person4.user, permissions: []},
        {user: person5.user, permissions: [viewPerms.permission2]}
      ]

      return Promise.each(tests, function (test) {
        return Permission.whereUser(test.user, {
          privilege: viewPerms.name
        }).then(function (permissions) {
          var ids = _.pluck(permissions, 'id')
          var expectedIds = _.pluck(test.permissions, 'id')
          expect(ids).to.eql(expectedIds)
        })
      })
    })

    it('scope by organization', () => {
      var tests = [
        {user: person.user, permissions: [editOrg.permission, viewPerms.permission]},
        {user: person2.user, permissions: [viewPerms.permission, editPerms.permission]},
        {user: person3.user, permissions: []},
        {user: person4.user, permissions: []},
        {user: person4.user, permissions: []}
      ]

      return Promise.each(tests, function (test) {
        return Permission.whereUser(test.user, {
          organizationId: organization.id
        }).then(function (permissions) {
          var ids = _.pluck(permissions, 'id')
          var expectedIds = _.pluck(test.permissions, 'id')
          expect(ids).to.eql(expectedIds)
        })
      })
    })

    it('scope by resource', () => {
      var tests = [
        {user: person.user, permissions: []},
        {user: person2.user, permissions: [editPerms.permission]},
        {user: person3.user, permissions: []},
        {user: person4.user, permissions: []},
        {user: person5.user, permissions: []}
      ]

      return Promise.each(tests, function (test) {
        return Permission.whereUser(test.user, {
          privilege: editPerms.name,
          organization: organization.id,
          resourceType: 'permissions',
          resourceId: editOrg.id
        }).then(function (permissions) {
          var ids = _.pluck(permissions, 'id')
          var expectedIds = _.pluck(test.permissions, 'id')
          expect(ids).to.eql(expectedIds)
        })
      })
    })
  })

  describe('getPrivileges', () => {
    it('should return organization privilege names', () => {
      var tests = [
        {user: person.user, privileges: [editOrg, viewPerms]},
        {user: person2.user, privileges: [viewPerms]},
        {user: person3.user, privileges: []},
        {user: person4.user, privileges: []},
        {user: person5.user, privileges: []}
      ]

      return Promise.each(tests, function (test) {
        return Permission.getPrivileges(test.user, {
          organizationId: organization.id,
          resourceType: null
        }).then(function (privileges) {
          var expectedNames = _.pluck(test.privileges, 'name')
          expect(privileges).to.eql(expectedNames)
        })
      })
    })

    it('should return resource privilege names', () => {
      var tests = [
        {user: person.user, privileges: []},
        {user: person2.user, privileges: [editPerms]},
        {user: person3.user, privileges: []},
        {user: person4.user, privileges: []},
        {user: person5.user, privileges: []}
      ]

      return Promise.each(tests, function (test) {
        return Permission.getPrivileges(test.user, {
          organizationId: organization.id,
          resourceType: 'permissions'
        }).then(function (privileges) {
          var expectedNames = _.pluck(test.privileges, 'name')
          expect(privileges).to.eql(expectedNames)
        })
      })
    })
  })

  describe('createBy @cleandb', () => {
    it('should create a single permission', () => {
      return Permission.createBy(
        {privilege: editOrg.name, organizationId: organization2.id, expression: 'O' + organization2.id}
      ).then(function (permissions) {
        expect(permissions).to.have.length(1)

        var permission = permissions[0]
        expect(permission.privilegeId).to.eql(editOrg.id)
        expect(permission.resourceType).to.be.null
        expect(permission.resourceId).to.be.null
      })
    })

    it('should create multiple permissions', () => {
      return Permission.createBy([
        {privilege: editOrg.name, organizationId: organization2.id, expression: 'O' + organization2.id},
        {privilege: viewPerms.name, organizationId: organization2.id, expression: 'O' + organization2.id, resourceType: 'permissions'},
        {privilege: viewPerms.name, organizationId: organization2.id, expression: 'O' + organization2.id, resourceType: 'permissions', resourceId: 567}
      ]).then(function (permissions) {
        expect(permissions).to.have.length(3)

        let permission = permissions[0]
        expect(permission.privilegeId).to.eql(editOrg.id)
        expect(permission.resourceType).to.be.null
        expect(permission.resourceId).to.be.null

        permission = permissions[1]
        expect(permission.privilegeId).to.eql(viewPerms.id)
        expect(permission.resourceType).to.eql('permissions')
        expect(permission.resourceId).to.be.null

        permission = permissions[2]
        expect(permission.privilegeId).to.eql(viewPerms.id)
        expect(permission.resourceType).to.eql('permissions')
        expect(permission.resourceId).to.eql(567)
      })
    })
  })

  describe('schema', () => {
    describe('loadPrivileges', () => {
      it('should load privilege names on the organization', () => {
        var tests = [
          {user: person.user, privileges: [editOrg, viewPerms]},
          {user: person2.user, privileges: [viewPerms]},
          {user: person3.user, privileges: []},
          {user: person4.user, privileges: []},
          {user: person5.user, privileges: []}
        ]

        return Promise.each(tests, function (test) {
          return Organization.loadPrivileges(organization, test.user).then(function (permission) {
            var expectedNames = _.pluck(test.privileges, 'name')
            expect(permission.privileges).to.eql(expectedNames)
          })
        })
      })

      it('should load privilege names on resources', () => {
        var tests = [
          {user: person.user, privileges: []},
          {user: person2.user, privileges: [editPerms]},
          {user: person3.user, privileges: []},
          {user: person4.user, privileges: []},
          {user: person5.user, privileges: []}
        ]

        return Promise.each(tests, function (test) {
          return Permission.loadPrivileges(editOrg, test.user).then(function (permission) {
            var expectedNames = _.pluck(test.privileges, 'name')
            expect(permission.privileges).to.eql(expectedNames)
          })
        })
      })
    })
  })

  describe('queries @cleandb', () => {
    describe('whereUserCan', () => {
      it('should restrict results by permissions', () => {
        var tests = [
          {user: person.user, permissions: []},
          {user: person2.user, permissions: [editOrg.permission]},
          {user: person3.user, permissions: []},
          {user: person4.user, permissions: []},
          {user: person5.user, permissions: []}
        ]

        return Promise.each(tests, function (test) {
          return Permission.whereUserCan(test.user, 'Edit Permissions').then(function (permissions) {
            var ids = _.pluck(permissions, 'id')
            var expectedIds = _.pluck(test.permissions, 'id')
            expect(ids).to.eql(expectedIds)
          })
        })
      })
    })
  })
})
