/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const Promise = require('bluebird')
const fs = require('fs')
const path = require('path')
const {expect} = require('chai')
const Factory = require('../factory')
const knex = require('../knex')
const {Group, Organization, Permission, Person} = require('../rbac')

describe('Person', () => {
  var organization
  var organization2
  var organization3
  var group
  var person
  var person2
  var person3
  var person4
  var person5

  beforeEach(async () => {
    organization = await Factory.create('organization')
    organization2 = await Factory.create('organization')
    organization3 = await Factory.create('organization')

    person = await Factory.createPerson(organization, {email: 'sam@test.com', fullName: 'Sam Tacoson', member: true, employeeId: '12345'})
    person2 = await Factory.createPerson(organization, {member: true})
    person3 = await Factory.createPerson(organization)
    person4 = await Factory.createPerson(organization, {member: true, archived: true})
    person5 = await Factory.createPerson(organization2, {member: true})

    group = await Factory.createGroup(organization, {name: 'Admin'})
    await Factory.addGroupMember(group, person)
    await Factory.addGroupMember(group, person3)
    await Factory.addGroupMember(group, person4)
  })

  describe('findByEmailId', () => {
    var user
    var member
    var friend
    var archived

    beforeEach(async () => {
      user = await Factory.createUser()
      archived = await Factory.createPerson(organization, {email: user.email, member: true, archived: true})
      friend = await Factory.createPerson(organization2, {email: user.email})
      member = await Factory.createPerson(organization3, {email: user.email, member: true})
    })

    describe('emailId only', () => {
      it('should find member', () => {
        return Person.findByEmailId(user.emailId).then(function (p) {
          expect(p.id).to.eql(member.id)
        })
      })

      it('should NOT find person with WRONG email', () => {
        return Person.findByEmailId(99999).then(function (p) {
          expect(p).to.be.undefined
        })
      })
    })

    describe('emailId and organization', () => {
      it('should find member', () => {
        return Person.findByEmailId(user.emailId, organization3).then(function (p) {
          expect(p.id).to.eql(member.id)
        })
      })

      it('should find friend', () => {
        return Person.findByEmailId(user.emailId, organization2).then(function (p) {
          expect(p.id).to.eql(friend.id)
        })
      })

      it('should find archived', () => {
        return Person.findByEmailId(user.emailId, organization).then(function (p) {
          expect(p.id).to.eql(archived.id)
        })
      })

      it('should NOT find person by WRONG organization', () => {
        return Person.findByEmailId(person2.emailId, organization2).then(function (p) {
          expect(p).to.be.undefined
        })
      })

      it('should NOT find person by WRONG emailId', () => {
        return Person.findByEmailId(99999, organization).then(function (p) {
          expect(p).to.be.undefined
        })
      })
    })
  })

  describe('findByEmail', () => {
    var user
    var member
    var friend
    var archived

    beforeEach(async () => {
      user = await Factory.createUser()
      archived = await Factory.createPerson(organization, {email: user.email, member: true, archived: true})
      friend = await Factory.createPerson(organization2, {email: user.email})
      member = await Factory.createPerson(organization3, {email: user.email, member: true})
    })

    describe('email only', () => {
      it('should find person', () => {
        return Person.findByEmail(user.email).then(function (p) {
          expect(p.id).to.eql(member.id)
        })
      })

      it('should NOT find person with WRONG email', () => {
        return Person.findByEmail('wrong@wrong.com').then(function (p) {
          expect(p).to.be.undefined
        })
      })
    })

    describe('email and organization', () => {
      it('should find member', () => {
        return Person.findByEmail(user.email, organization3).then(function (p) {
          expect(p.id).to.eql(member.id)
        })
      })

      it('should find archived', () => {
        return Person.findByEmail(user.email, organization2).then(function (p) {
          expect(p.id).to.eql(friend.id)
        })
      })

      it('should find archived', () => {
        return Person.findByEmail(user.email, organization).then(function (p) {
          expect(p.id).to.eql(archived.id)
        })
      })

      it('should NOT find person by WRONG organization', () => {
        return Person.findByEmail(person2.email, organization2).then(function (p) {
          expect(p).to.be.undefined
        })
      })

      it('should NOT find person by WRONG emailId', () => {
        return Person.findByEmail('wrong@wrong.com', organization).then(function (p) {
          expect(p).to.be.undefined
        })
      })
    })
  })

  describe('whereSearch @cleandb', () => {
    it('should find people with search text (by name)', () => {
      return Person.query().whereSearch('sa').then(function (users) {
        var names = _.pluck(users, 'fullName')
        expect(names).to.eql(['Sam Tacoson'])
      })
    })

    it('should find people with search text (by email)', () => {
      return Person.query().whereSearch('taco').then(function (users) {
        var names = _.pluck(users, 'fullName')
        expect(names).to.eql(['Sam Tacoson'])
      })
    })

    it('should find people with search text (by id)', () => {
      return Person.query().whereSearch('345').then(function (users) {
        var names = _.pluck(users, 'fullName')
        expect(names).to.eql(['Sam Tacoson'])
      })
    })
  })

  describe('whereExpression', () => {
    describe('raw query', () => {
      it('should match an empty expression', () => {
        var expression = ''
        return Person.query().whereExpression(expression).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.be.empty
        })
      })

      it('should match a Person expression', () => {
        var expression = 'P' + person.id
        return Person.query().whereExpression(expression).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id])
        })
      })

      it('should match a Group expression', () => {
        var expression = 'G' + group.id
        return Person.query().whereExpression(expression).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id, person3.id, person4.id])
        })
      })

      it('should match a Organization expression', () => {
        var expression = 'O' + organization.id
        return Person.query().whereExpression(expression).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id, person2.id, person4.id])
        })
      })

      it('should match a negation expression @cleandb', () => {
        var expression = '!P' + person.id
        return Person.query().whereExpression(expression).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person2.id, person3.id, person4.id, person5.id])
        })
      })

      it('should match a conjunction expression', () => {
        var expression = 'O' + organization.id + '||' + 'G' + group.id
        return Person.query().whereExpression(expression).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id, person2.id, person3.id, person4.id])
        })
      })
    })

    describe('within organizational bounds', () => {
      it('should match an empty expression', () => {
        var expression = ''
        return Person.whereExpression(expression, organization.id).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.be.empty
        })
      })

      it('should match a Person expression', () => {
        var expression = 'P' + person.id
        return Person.whereExpression(expression, organization.id).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id])
        })
      })

      it('should match a Group expression', () => {
        var expression = 'G' + group.id
        return Person.whereExpression(expression, organization.id).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id, person3.id])
        })
      })

      it('should match a Organization expression', () => {
        var expression = 'O' + organization.id
        return Person.whereExpression(expression, organization.id).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id, person2.id])
        })
      })

      it('should match a negation expression', () => {
        var expression = '!P' + person.id
        return Person.whereExpression(expression, organization.id).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person2.id, person3.id])
        })
      })

      it('should match a conjunction expression', () => {
        var expression = 'O' + organization.id + '||' + 'G' + group.id
        return Person.whereExpression(expression, organization.id).then(function (people) {
          var ids = _.pluck(people, 'id')
          expect(ids).to.eql([person.id, person2.id, person3.id])
        })
      })
    })
  })

  describe('whereQuery', () => {
    it('should only find people on the organization (members)', () => {
      return Person.query().whereQuery({organization: organization.id, member: 'true'}).select('people.*').then(function (people) {
        var ids = _.pluck(people, 'id')
        expect(ids).to.eql([person.id, person2.id])
      })
    })

    it('should only find people on the organization (members & friends)', () => {
      return Person.query().whereQuery({organization: organization.id}).select('people.*').then(function (people) {
        var ids = _.pluck(people, 'id')
        expect(ids).to.eql([person.id, person2.id, person3.id])
      })
    })

    it('should only find people on the organization (friends only)', () => {
      return Person.query().whereQuery({organization: organization.id, member: 'false'}).select('people.*').then(function (people) {
        var ids = _.pluck(people, 'id')
        expect(ids).to.eql([person3.id])
      })
    })

    it('should return archived people', () => {
      return Person.query().whereQuery({organization: organization.id, archived: 'true'}).select('people.*').then(function (people) {
        var ids = _.pluck(people, 'id')
        expect(ids).to.eql([person.id, person2.id, person3.id, person4.id])
      })
    })

    it('should return people IN group', () => {
      return Person.whereOrganizationId(organization.id).whereQuery({group: group.id}).select('people.*').then(function (people) {
        var ids = _.pluck(people, 'id')
        expect(ids).to.eql([person.id, person3.id])
      })
    })

    it('should return people NOT IN group', () => {
      return Person.whereOrganizationId(organization.id).whereQuery({group: -group.id}).select('people.*').then(function (people) {
        var ids = _.pluck(people, 'id')
        expect(ids).to.eql([person2.id])
      })
    })
  })

  describe('findOrCreateBy', () => {
    it('should find an existing person', () => {
      return Person.findOrCreateBy({
        organizationId: organization.id,
        email: person.email
      }).then(function (p) {
        expect(p.id).to.eql(person.id)
        expect(p.email).to.eql(person.email)
        expect(p.fullName).to.eql(person.fullName)
      })
    })

    it('should create a person', () => {
      return Person.findOrCreateBy({
        organizationId: organization.id,
        email: 'test@test.com',
        fullName: 'Tester'
      }).then(function (p) {
        expect(p.id).to.not.be.undefined
        expect(p.email).to.eql('test@test.com')
        expect(p.fullName).to.eql('Tester')
      })
    })

    it('should create a person (transaction)', () => {
      var email = 'test@test.com'
      return knex.transaction(function (trx) {
        return Person.findOrCreateBy({
          organizationId: organization.id,
          email: email,
          fullName: 'Tester'
        }, trx).then(function (p) {
          expect(p.id).to.not.be.undefined
          expect(p.email).to.eql(email)
          expect(p.fullName).to.eql('Tester')
        }).then(trx.commit)
      }).then(() => {
        return Person.findByEmail(email, organization)
      }).then(function (person) {
        expect(person).to.not.be.undefined
      })
    })

    it('should create a person (and rollback)', () => {
      return knex.transaction(function (trx) {
        return Person.findOrCreateBy({
          organizationId: organization.id,
          email: 'test@test.com',
          fullName: 'Tester'
        }, trx).then(function (p) {
          expect(p.id).to.not.be.undefined
          expect(p.email).to.eql('test@test.com')
          expect(p.fullName).to.eql('Tester')
        }).then(trx.rollback)
      }).catch(function () {
        return Person.findByEmail('test@test.com', organization)
      }).then(function (person) {
        expect(person).to.be.undefined
      })
    })
  })

  describe('updateOrCreateBy', () => {
    it('should update an existing person', () => {
      return Person.updateOrCreateBy({
        organizationId: organization.id,
        email: person.email,
        fullName: 'New Name'
      }).then(function (p) {
        expect(p.id).to.eql(person.id)
        expect(p.email).to.eql(person.email)
        expect(p.fullName).to.eql('New Name')
      })
    })

    it('should create a person', () => {
      return Person.updateOrCreateBy({
        organizationId: organization.id,
        email: 'test@test.com',
        fullName: 'Tester'
      }).then(function (p) {
        expect(p.id).to.not.be.undefined
        expect(p.email).to.eql('test@test.com')
        expect(p.fullName).to.eql('Tester')
      })
    })

    it('should create a person (transaction)', () => {
      var email = 'test@test.com'
      return knex.transaction(function (trx) {
        return Person.updateOrCreateBy({
          organizationId: organization.id,
          email: email,
          fullName: 'Tester'
        }, trx).then(function (p) {
          expect(p.id).to.not.be.undefined
          expect(p.email).to.eql(email)
          expect(p.fullName).to.eql('Tester')
        }).then(trx.commit)
      }).then(() => {
        return Person.findByEmail(email, organization)
      }).then(function (person) {
        expect(person).to.not.be.undefined
      })
    })

    it('should create a person (and rollback)', () => {
      return knex.transaction(function (trx) {
        return Person.updateOrCreateBy({
          organizationId: organization.id,
          email: 'test@test.com',
          fullName: 'Tester'
        }, trx).then(function (p) {
          expect(p.id).to.not.be.undefined
          expect(p.email).to.eql('test@test.com')
          expect(p.fullName).to.eql('Tester')
        }).then(trx.rollback)
      }).catch(function () {
        return Person.findByEmail('test@test.com', organization)
      }).then(function (person) {
        expect(person).to.be.undefined
      })
    })
  })

  describe('bulkLoad', () => {
    var tests = [{
      name: 'throw an error if duplicate employee ids',
      entries: [
        {employeeId: 1, email: 'a@a.com', fullName: 'Frank'},
        {employeeId: 1, email: 'b@b.com', fullName: 'Tacos'}
      ],
      error: Person.bulkError
    }, {
      name: 'throw an error if duplicate emails',
      entries: [
        {employeeId: 1, email: 'a@a.com', fullName: 'Frank'},
        {employeeId: 2, email: 'a@a.com', fullName: 'Tacos'}
      ],
      error: Person.bulkError
    }]

    _.each(tests, function (test) {
      it(`should ${test.name}`, () => {
        var bulkLoad = _.partial(Person.bulkLoad, organization, test.entries)
        expect(bulkLoad).to.throw(Person.bulkError)
      })
    })
  })

  describe('bulkCreate', () => {
    var newMember = {email: 'member@member.com', fullName: 'Member #1', groups: ['Admin', 'DC'], member: true}
    var newMember2 = {email: 'member2@member.com', fullName: 'Member #2', groups: ['GA'], member: true}

    it('should unarchive archived person', () => {
      var params = _.pick(person4, 'email', 'fullName')
      return Person.bulkCreate(organization, params).then(function (people) {
        var emails = _.pluck(people, 'email')
        expect(emails).to.eql([person4.email])

        return Person.where({organizationId: organization.id}).joinEmail().then(function (people) {
          var members = _.where(people, {member: true, archived: false})
          var emails = _.pluck(members, 'email')
          expect(emails).to.eql([person.email, person2.email, person4.email])
        })
      })
    })

    _.each(['email', 'employeeId'], function (testField) {
      it(`should alter existing person found via ${testField}`, () => {
        var params = _.pick(person, testField)
        params.fullName = 'New Name'
        return Person.bulkCreate(organization, params).then(function (results) {
          // Should be same person
          var result = _.first(results)
          expect(result.id).to.eql(person.id)
          expect(result.fullName).to.eql('New Name')
          return Group.wherePersonId(person.id)
        }).then(function (groups) { // Should be same groups
          var names = _.pluck(groups, 'name')
          expect(names).to.eql(['Admin'])
        })
      })
    })

    it('should create a new member', () => {
      return Person.bulkCreate(organization, newMember).then(function (people) {
        var newPerson = _.first(people)
        var emails = _.pluck(people, 'email')
        expect(emails).to.eql([newMember.email])

        return Promise.all([
          Person.where({organizationId: organization.id}).joinEmail(),
          Organization.loadGroups(organization)
        ]).spread(function (people, organization) {
          // Members
          var members = _.where(people, {member: true, archived: false})
          var emails = _.pluck(members, 'email')
          expect(emails).to.eql([person.email, person2.email, newMember.email])
          // Group Names
          var names = _.chain(organization.groups).pluck('name').uniq().value()
          expect(names).to.eql(['Admin', 'DC'])
          // Admin Group Members
          var admin = _.findWhere(organization.groups, {name: 'Admin'})
          expect(admin.peopleCount).to.eql(3)
          return Group.Member.whereGroupId(admin.id)
        }).then(function (members) {
          var personIds = _.pluck(members, 'personId')
          expect(personIds).to.eql([person.id, person3.id, person4.id, newPerson.id])
        })
      })
    })

    it('should create new members', () => {
      var oldMember = _.pick(person, 'email', 'fullName')
      oldMember.groups = ['DC']
      return Person.bulkCreate(organization, [oldMember, newMember, newMember2]).then(function (people) {
        var newPerson = _.findWhere(people, {email: newMember.email})
        var emails = _.pluck(people, 'email')
        expect(emails).to.eql([person.email, newMember.email, newMember2.email])

        return Promise.all([
          Person.where({organizationId: organization.id}).joinEmail(),
          Organization.loadGroups(organization)
        ]).spread(function (people, organization) {
          // Members
          var members = _.where(people, {member: true, archived: false})
          var emails = _.pluck(members, 'email')
          expect(emails).to.eql([person.email, person2.email, newMember.email, newMember2.email])
          // Group Names
          var names = _.chain(organization.groups).pluck('name').uniq().value()
          expect(names).to.eql(['Admin', 'DC', 'GA'])
          // Admin Group Members
          var admin = _.findWhere(organization.groups, {name: 'Admin'})
          expect(admin.peopleCount).to.eql(2)
          return Group.Member.whereGroupId(admin.id)
        }).then(function (members) {
          var personIds = _.pluck(members, 'personId')
          expect(personIds).to.eql([person3.id, person4.id, newPerson.id])
          // DC Group Members
          var dc = _.findWhere(organization.groups, {name: 'DC'})
          expect(dc.peopleCount).to.eql(2)
          return Group.Member.whereGroupId(dc.id)
        }).then(function (members) {
          var personIds = _.pluck(members, 'personId')
          expect(personIds).to.eql([person.id, newPerson.id])
        })
      })
    })
  })

  describe('archive', () => {
    var expression
    var privilege
    var permission

    beforeEach(async () => {
      expression = 'O' + organization.id + '||' + 'P' + person3.id
      privilege = await Factory.create('privilege')
      permission = await Factory.createPermission(privilege, organization, expression)
    })

    it('archive a member from the organization ONLY', () => {
      return Person.archive(person2).then(() => {
        expect(person2.archived).to.be.true
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      }).then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person.email])

        // Archived
        members = _.where(people, {archived: true})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person2.email, person4.email])

        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id, person3.id, person4.id])
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql(expression)
      })
    })

    it('archive a member from the organization AND groups', () => {
      return Person.archive(person).then(() => {
        expect(person.archived).to.be.true
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      }).then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person2.email])

        // Archived
        members = _.where(people, {archived: true})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person.email, person4.email])

        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id, person3.id, person4.id])
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql(expression)
      })
    })

    it('archive a member from the groups ONLY', () => {
      return Person.archive(person3).then(() => {
        expect(person3.archived).to.be.true
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      }).then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person.email, person2.email])

        // Archived
        members = _.where(people, {archived: true})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person3.email, person4.email])

        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id, person3.id, person4.id])
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql('O' + organization.id)
      })
    })
  })

  describe('unarchive', () => {
    it('unarchive a member', () => {
      return Person.unarchive(person4).then(() => {
        expect(person4.archived).to.be.false
      })
    })
  })

  describe('purge', () => {
    var expression
    var privilege
    var permission

    beforeEach(async () => {
      expression = 'O' + organization.id + '||' + 'P' + person3.id
      privilege = await Factory.create('privilege')
      permission = await Factory.createPermission(privilege, organization, expression)
    })

    it('remove a member from the organization ONLY', () => {
      return Person.purge(person2).then(() => {
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      }).then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person.email])

        // Archived
        members = _.where(people, {archived: true})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person4.email])

        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id, person3.id, person4.id])
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql(expression)
      })
    })

    it('remove a member from the organization AND groups', () => {
      return Person.purge(person).then(() => {
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      }).then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person2.email])

        // Archived
        members = _.where(people, {archived: true})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person4.email])

        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person3.id, person4.id])
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql(expression)
      })
    })

    it('remove a member from the groups ONLY', () => {
      return Person.purge(person3).then(() => {
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      }).then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person.email, person2.email])

        // Archived
        members = _.where(people, {archived: true})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql([person4.email])

        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id, person4.id])
        return Permission.findById(permission.id)
      }).then(function (permission) {
        expect(permission.expression).to.eql('O' + organization.id)
      })
    })
  })

  describe('bulkUpload', () => {
    function createStream (filename) {
      var csv = path.resolve(__dirname, '../data', filename)
      return fs.createReadStream(csv)
    }

    function bulkUpload () {
      var filenames = _.map(arguments, _.identity)
      return Promise.each(filenames, function (filename) {
        var stream = createStream(filename)
        return Person.bulkUpload(organization, stream, person.user)
      }).then(() => {
        return Person.query().joinEmail().whereOrganizationId(organization.id)
      })
    }

    it('should catch an empty file', () => {
      return bulkUpload('members-empty.csv').catch(function (err) {
        expect(err.message).to.eql('The file provided is empty')
      })
    })

    it('should catch a file with no emails', () => {
      return bulkUpload('members-no-emails.csv').catch(function (err) {
        expect(err.message).to.eql('Missing required column: Email')
      })
    })

    it('should catch a file with a missing email', () => {
      return bulkUpload('members-missing-email.csv').catch(function (err) {
        expect(err.message).to.eql('Missing Email on line 3')
      })
    })

    it('should catch a file with a missing name', () => {
      return bulkUpload('members-missing-name.csv').catch(function (err) {
        expect(err.message).to.eql('Missing Name on line 3')
      })
    })

    it('should catch a file without my email', () => {
      return bulkUpload('members-missing-me.csv').catch(function (err) {
        expect(err.message).to.eql('Could not find your email address: ' + person.email)
      })
    })

    it('should catch a file with duplicate emails', () => {
      return bulkUpload('members-duplicate-emails.csv').catch(function (err) {
        expect(err.message).to.eql('Duplicate Email: maria@test.com')
      })
    })

    it('should catch a file with duplicate ids', () => {
      return bulkUpload('members-duplicate-ids.csv').catch(function (err) {
        expect(err.message).to.eql('Duplicate ID: 1')
      })
    })

    it('should load new members', () => {
      return bulkUpload('members.csv').then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql(['sam@test.com', 'maria@test.com'])

        // Archived
        let archived = _.where(people, {archived: true})
        emails = _.pluck(archived, 'email')
        expect(emails).to.eql([person2.email, person4.email])

        return Organization.loadGroups(organization)
      }).then(function (organization) {
        // Should be on the new group
        var group = _.findWhere(organization.groups, {name: 'Group #1'})
        expect(group.peopleCount).to.eql(1)
        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id])
        // Should still be on the admin group
        var group = _.findWhere(organization.groups, {name: 'Admin'})
        expect(group.peopleCount).to.eql(1)
        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person3.id, person4.id])
      })
    })

    it('should remove old members', () => {
      return bulkUpload('members.csv', 'members-remove.csv').then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql(['sam@test.com'])

        // Archived
        let archived = _.where(people, {archived: true})
        emails = _.pluck(archived, 'email')
        expect(emails).to.eql([person2.email, person4.email, 'maria@test.com'])
      })
    })

    it('should revive archived members', () => {
      return bulkUpload('members.csv', 'members-remove.csv', 'members.csv').then(function (people) {
        // Friends
        let friends = _.where(people, {member: false, archived: false})
        let emails = _.pluck(friends, 'email')
        expect(emails).to.eql([person3.email])

        // Members
        let members = _.where(people, {member: true, archived: false})
        emails = _.pluck(members, 'email')
        expect(emails).to.eql(['sam@test.com', 'maria@test.com'])

        // Archived
        let archived = _.where(people, {archived: true})
        emails = _.pluck(archived, 'email')
        expect(emails).to.eql([person2.email, person4.email])
      })
    })

    it('should remove group members', () => {
      return bulkUpload('members.csv', 'members-remove-group.csv').then(function (people) {
        var person = _.findWhere(people, {email: 'sam@test.com'})
        return Group.Member.wherePersonId(person.id)
      }).then(function (members) {
        expect(members).to.be.empty
      })
    })

    it('should add employee ids', () => {
      return bulkUpload('members-id.csv').then(function (people) {
        let members = _.where(people, {member: true, archived: false})
        expect(members).to.have.length(2)

        let person = _.findWhere(members, {email: 'sam@test.com'})
        expect(person.employeeId).to.eql('1')

        person = _.findWhere(members, {email: 'maria@test.com'})
        expect(person.employeeId).to.eql('2')
      })
    })

    it('should modify a name', () => {
      var oldMembers
      return bulkUpload('members.csv').then(function (people) {
        oldMembers = _.where(people, {member: true})
        return bulkUpload('members-change-name.csv')
      }).then(function (people) {
        var members = _.where(people, {member: true})

        var oldPerson = _.findWhere(oldMembers, {email: 'maria@test.com'})
        var person = _.findWhere(members, {email: 'maria@test.com'})
        expect(person.fullName).to.eql('Maria-Teresa Duvall')
        expect(person.id).to.eql(oldPerson.id)
      })
    })

    it('should modify an email address', () => {
      var oldMembers
      return bulkUpload('members-id.csv').then(function (people) {
        oldMembers = _.where(people, {member: true})
        return bulkUpload('members-id-change-email.csv')
      }).then(function (people) {
        var members = _.where(people, {member: true})
        var oldPerson = _.findWhere(oldMembers, {email: 'maria@test.com'})
        var person = _.findWhere(members, {email: 'maria.duvall@test.com'})
        expect(person.employeeId).to.eql('2')
        expect(person.id).to.eql(oldPerson.id)
      })
    })

    it('should not modify groups, if not present', () => {
      return bulkUpload('members-no-groups.csv').then(function (people) {
        return Organization.loadGroups(organization)
      }).then(() => {
        var group = _.findWhere(organization.groups, {name: 'Admin'})
        expect(group.peopleCount).to.eql(2)
        return Group.Member.whereGroupId(group.id)
      }).then(function (members) {
        var personIds = _.pluck(members, 'personId')
        expect(personIds).to.eql([person.id, person3.id, person4.id])
      })
    })
  })

  describe('bulkMerge', () => {
    var tests = [{
      name: 'leave well enough alone',
      before: [
        {email: 'tom@tom.com', fullName: 'Tom'},
        {email: 'sam@sam.com', fullName: 'Sam'}
      ],
      after: [
        {email: 'tom@tom.com', fullName: 'Tom'},
        {email: 'sam@sam.com', fullName: 'Sam'}
      ]
    }, {
      name: 'merge groups',
      before: [
        {email: 'tom@tom.com', fullName: 'Tom', groups: ['1']},
        {email: 'sam@sam.com', fullName: 'Sam', groups: ['2']},
        {email: 'tom@tom.com', fullName: 'Tommy', groups: ['3']}
      ],
      after: [
        {email: 'tom@tom.com', fullName: 'Tom', groups: ['1', '3']},
        {email: 'sam@sam.com', fullName: 'Sam', groups: ['2']}
      ]
    }, {
      name: 'merge employee id / email',
      before: [
        {employeeId: 1, email: 'tom@tom.com', fullName: 'Tom', groups: ['1']},
        {employeeId: 2, email: 'sam@sam.com', fullName: 'Sam', groups: ['2']},
        {employeeId: 1, email: 'tom@tom.com', fullName: 'Tommy', groups: ['3']},
        {employeeId: 3, email: undefined, fullName: 'Tommy', groups: ['4']},
        {employeeId: undefined, email: 'tom@tom.com', fullName: 'Tommy', groups: ['5']}
      ],
      after: [
        {employeeId: 1, email: 'tom@tom.com', fullName: 'Tom', groups: ['1', '3']},
        {employeeId: 2, email: 'sam@sam.com', fullName: 'Sam', groups: ['2']},
        {employeeId: 3, email: undefined, fullName: 'Tommy', groups: ['4']},
        {employeeId: undefined, email: 'tom@tom.com', fullName: 'Tommy', groups: ['5']}
      ]
    }]

    _.each(tests, function (test) {
      it('should ' + test.name, () => {
        var result = Person.bulkMerge(test.before)
        expect(result).to.eql(test.after)
      })
    })
  })

  describe('getVariants', () => {
    var organization
    var group
    var person
    var person2
    var person3
    var person4

    beforeEach(async () => {
      organization = await Factory.create('organization')
      person = await Factory.createPerson(organization, {fullName: 'Sam', member: true})
      person2 = await Factory.createPerson(organization, {fullName: 'Tom', member: true})
      person3 = await Factory.createPerson(organization, {fullName: 'Maria'})
      person4 = await Factory.createPerson(organization, {fullName: 'Locke'})

      group = await Factory.createGroup(organization)
      await Factory.addGroupMember(group, person)
      await Factory.addGroupMember(group, person3)
    })

    it('should map organizations, groups and people', () => {
      var tests = [
        {person: person, values: ['O' + organization.id, 'G' + group.id, 'P' + person.id]},
        {person: person2, values: ['O' + organization.id, 'P' + person2.id]},
        {person: person3, values: ['G' + group.id, 'P' + person3.id]},
        {person: person4, values: ['P' + person4.id]}
      ]

      return Promise.each(tests, function (test) {
        return Person.getVariants(test.person).then(function (variants) {
          var values = _.invoke(variants, 'toString')
          expect(values).to.eql(test.values)
        })
      })
    })
  })
})
