/* global beforeEach describe it */
/* eslint-disable no-unused-expressions */
const _ = require('underscore')
const moment = require('moment')
const {expect} = require('chai')
const Factory = require('../factory')
const {User} = require('../rbac')

describe('User', () => {
  var user
  var email

  beforeEach(async () => {
    user = await Factory.createUser()
    email = await Factory.createEmail()
  })

  describe('findBy', () => {
    it('should NOT find a user with no parameters', () => {
      return User.findBy().then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should NOT find a user with empty parameters', () => {
      return User.findBy({}).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should NOT find an unknown user by emailId', () => {
      return User.findBy({emailId: email.id}).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should NOT find an unknown user by email', () => {
      return User.findBy({emailId: email.email}).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should find a known user by emailId', () => {
      return User.findBy({emailId: user.emailId}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql(user.fullName)
      })
    })

    it('should find a known user by email', () => {
      return User.findBy({email: user.email}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql(user.fullName)
      })
    })
  })

  describe('findOrBuildBy', () => {
    it('should NOT find a user with no parameters', () => {
      return User.findOrBuildBy().then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should NOT find a user with empty parameters', () => {
      return User.findOrBuildBy({}).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should build an unknown user by emailId', () => {
      return User.findOrBuildBy({emailId: email.id}).then(function (u) {
        expect(u.id).to.be.undefined
        expect(u.emailId).to.eql(email.id)
        expect(u.email).to.eql(email.email)
        expect(u.fullName).to.be.undefined
      })
    })

    it('should find a known user emailId', () => {
      return User.findOrBuildBy({emailId: user.emailId}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql(user.fullName)
      })
    })

    it('should find a known user by email', () => {
      return User.findOrBuildBy({email: user.email}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql(user.fullName)
      })
    })
  })

  describe('findOrCreateBy', () => {
    it('should NOT find a user with no parameters', () => {
      return User.findOrCreateBy().then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should NOT find a user with empty parameters', () => {
      return User.findOrCreateBy({fullName: 'Test'}).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should create an unknown user by emailId', () => {
      return User.findOrCreateBy({emailId: email.id, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.not.be.undefined
        expect(u.emailId).to.eql(email.id)
        expect(u.email).to.eql(email.email)
        expect(u.fullName).to.eql('Test Person')
      })
    })

    it('should create an unknown user by email', () => {
      return User.findOrCreateBy({email: email.email, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.not.be.undefined
        expect(u.emailId).to.eql(email.id)
        expect(u.email).to.eql(email.email)
        expect(u.fullName).to.eql('Test Person')
      })
    })

    it('should find a known user emailId', () => {
      return User.findOrCreateBy({emailId: user.emailId, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql(user.fullName)
      })
    })

    it('should find a known user by email', () => {
      return User.findOrCreateBy({email: user.email, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql(user.fullName)
      })
    })
  })

  describe('updateOrCreateBy', () => {
    it('should NOT find a user with no parameters', () => {
      return User.updateOrCreateBy().then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should NOT find a user with empty parameters', () => {
      return User.updateOrCreateBy({fullName: 'Test Person'}).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should create an unknown user by emailId', () => {
      return User.updateOrCreateBy({emailId: email.id, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.not.be.undefined
        expect(u.emailId).to.eql(email.id)
        expect(u.email).to.eql(email.email)
        expect(u.fullName).to.eql('Test Person')
      })
    })

    it('should create an unknown user by email', () => {
      return User.updateOrCreateBy({email: email.email, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.not.be.undefined
        expect(u.emailId).to.eql(email.id)
        expect(u.email).to.eql(email.email)
        expect(u.fullName).to.eql('Test Person')
      })
    })

    it('should find a known user emailId', () => {
      return User.updateOrCreateBy({emailId: user.emailId, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql('Test Person')
      })
    })

    it('should find a known user by email', () => {
      return User.updateOrCreateBy({email: user.email, fullName: 'Test Person'}).then(function (u) {
        expect(u.id).to.eql(user.id)
        expect(u.emailId).to.eql(user.emailId)
        expect(u.email).to.eql(user.email)
        expect(u.fullName).to.eql('Test Person')
      })
    })
  })

  describe('addAuthToken', () => {
    beforeEach(() => {
      user.authToken = 'token'
    })

    it('should add without existing query params', () => {
      var url = User.addAuthToken(user, 'http://blissbook.com/handbooks')
      expect(url).to.eql('http://blissbook.com/handbooks?auth_token=token')
    })

    it('should add with existing query params', () => {
      var url = User.addAuthToken(user, 'http://blissbook.com/handbooks?other=value')
      expect(url).to.eql('http://blissbook.com/handbooks?other=value&auth_token=token')
    })
  })

  describe('findByAuthToken', () => {
    it('should not find the email (bad token)', () => {
      return User.resetAuthToken(user).then(() => {
        return User.findByAuthToken('wrong token')
      }).then(function (u) {
        expect(u).to.be.undefined
      })
    })

    it('should find the email', () => {
      return User.resetAuthToken(user).then(() => {
        return User.findByAuthToken(user.authToken)
      }).then(function (u) {
        expect(u.id).to.eql(user.id)
      })
    })
  })

  describe('isAuthTokenValid', () => {
    it('should not validate an old token', () => {
      var old = moment().subtract(16, 'minutes').toDate()
      var valid = User.isAuthTokenValid(old)
      expect(valid).to.be.false
    })

    it('should validate a recent token', () => {
      var now = moment().toDate()
      var valid = User.isAuthTokenValid(now)
      expect(valid).to.be.true
    })
  })

  describe('generateAccessParams', () => {
    it('should create a token and timestamp', () => {
      // Parse params
      var params = User.generateAccessParams(user)
      expect(params.access_timestamp).to.not.be.undefined
      expect(params.access_token).to.not.be.undefined

      // Correct
      let query = _.pick(params, 'access_id', 'access_token', 'access_timestamp')
      let userId = User.parseAccessParams(query, 60)
      expect(userId).to.eql(user.id)

      // Incorrect Secret
      query = _.pick(params, 'access_id', 'access_timestamp')
      query.access_token = 'incorrect'
      userId = User.parseAccessParams(query, 60)
      expect(userId).to.be.undefined

      // Incorrect Time
      query = _.pick(params, 'access_id', 'access_token', 'access_timestamp')
      userId = User.parseAccessParams(query, -1)
      expect(userId).to.be.undefined
    })
  })
})
